from enum import Enum
from qutrunk.sim.qusprout.rpcclient import QuSproutApiServer
from qutrunk.sim.qusprout.qusproutdata import ttypes as qusproutdata
from qutrunk.tools.read_qubox import get_qubox_setting
from .backend import Backend


class ExecType(Enum):
    """
    Init exec type for quantum circuit

    Attributes:
        ExecBySingleProcess: Execute by single process
        ExecByMpi: Execute by multiple processes
    """
    SingleProcess = 1
    Mpi = 2


class BackendQuSprout(Backend):
    """
    QuSprout: quamtum circuit simulator, provide multi-threaded OMP, multi node parallel MPI, GPU hardware acceleration.
    To use qusprout, make sure the network is connected and the service IP and Port are set correctly.
    """
    def __init__(self, exectype=ExecType.SingleProcess):
        super().__init__()
        self.circuit = None
        self.exectype = exectype
        box_config = get_qubox_setting()
        self._api_server = QuSproutApiServer(ip=box_config.get('ip'), port=box_config.get('port'))

    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        res, elapsed = self._api_server.get_prob_amp(index)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_all_outcome(self, qubits):
        """
        Get outcomeProbs with the probabilities of every outcome of the sub-register contained in qureg

        Args:
            qubits: the sub-register contained in qureg

        Returns:
            An array contains probability of target qubits
        """
        res, elapsed = self._api_server.get_prob_all_outcome(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_outcome(self, qubit, outcome):
        """
        Get the probability of a specified qubit being measured in the given outcome (0 or 1)

        Args:
            qubit: the specified qubit to be measured
            outcome: the qubit measure result(0 or 1)

        Returns:
            the probability of target qubit 
        """
        res, elapsed = self._api_server.get_prob_outcome(qubit, outcome)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        res, elapsed = self._api_server.get_all_state()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def send_circuit(self, circuit, final=False):
        """
        Send the quantum circuit to qusprout backend

        Args:
            circuit: quantum circuit to send
            final: True if quantum circuit finish, default False, \
                when final==True The backend program will release the computing resources 
        """
        cmds = []
        start = circuit.cmd_cursor
        stop = len(circuit.cmds)

        for idx in range(start, stop):
            cmd = circuit.cmds[idx]
            c = qusproutdata.Cmd(str(cmd.gate), cmd.targets, cmd.controls, cmd.rotation, cmd.qasm(), cmd.inverse)
            cmds.append(c)

        circuit.forward(stop - start)

        # 服务端初始化
        if start == 0:
            res, elapsed = self._api_server.init(
                circuit.qubits_len, 
                circuit.initstate, 
                circuit.initvalue,
                circuit.get_density(),
                self.exectype.value)
            if self.circuit.counter:
                self.circuit.counter.acc_run_time(elapsed)

        if len(cmds) == 0 and (not final):
            return

        # 发送至服务        
        res, elapsed = self._api_server.send_circuit(qusproutdata.Circuit(cmds), final)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)

    def run(self, shots=1):
        """
        Run quantum circuit

        Args:
            shots: circuit run times, for sampling, default: 1
        
        Returns:
            result: the Result object contain circuit running outcome
        """
        res, elapsed = self._api_server.run(shots)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
            self.circuit.counter.finish() 

        """
        1 必须释放连接，不然其它连接无法连上服务端
        2 不能放在__del__中，因为对象释放不代表析构函数会及时调用
        """
        self._api_server.close()

        return res

    def apply_QFT(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        self._api_server.apply_QFT(qubits)

    def apply_Full_QFT(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        self._api_server.apply_Full_QFT()

    def get_expec_pauli_prod(self, pauli_prod_list):
        """
        Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: a list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            the expected value of a product of Pauli operators.
        """
        puali_list = []
        for temp in pauli_prod_list:
            puali_list.append(qusproutdata.PauliProdInfo(temp['oper_type'], temp['target']))
            
        res, elapsed = self._api_server.get_expec_pauli_prod(puali_list)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res
    
    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: a list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: the coefficients of each term in the sum of Pauli products.

        Returns:
            the expected value of a sum of products of Pauli operators. 
        """
        res, elapsed = self._api_server.get_expec_pauli_sum(oper_type_list, term_coeff_list)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res



"""Circuit Module"""
import json
import random
import sys
from enum import Enum
from typing import List, Optional

from qutrunk.backends import Backend, BackendLocal
from qutrunk.circuit import CBit, CReg, QuBit, Qureg, Counter
from qutrunk.circuit.gates import MeasureGate
from qutrunk.circuit.ops import Observable


class InitState(Enum):
    """Init state for quantum circuit(represent by qureg).

    Attributes:
        Zero: Initialise qureg into the zero state
        Plus: Initialise qureg into the plus state.
        Classical: Initialise qureg into the classical state with index stateInd
    """

    ZERO = 1
    PLUS = 2
    CLASSICAL = 3


class QCircuit:
    """
    Quantum circuit

    Attributes:
        backend: Used to run quantum circuits
        density: Creates a density matrix Qureg object representing a set of \
            qubits which can enter noisy and mixed states
        name: The circuit name
        counter: Statistics the resources of circuit used
    """

    prefix = "circuit"

    def __init__(
            self,
            backend=None,
            density=False,
            name: Optional[str] = None,
            resource: Optional[bool] = False,
    ):
        self.qreg = None
        self.creg = None
        self.cmds = []
        self.init_state = None
        self.init_value = None
        self.cmd_cursor = 0
        self.counter = None

        self.qubit_indices = {}
        self.cbit_indices = {}

        # 默认使用local作为后端
        if backend is None:
            self.backend = BackendLocal()
        else:
            if not isinstance(backend, Backend):
                raise TypeError("You supplied a backend which is not supported.\n")
            self.backend = backend

        # 本地模式暂不支持噪声模型
        if density and isinstance(backend, BackendLocal):
            raise TypeError("You supplied a backend which is not supported density.\n")
        self.density = density

        self.backend.circuit = self
        self.outcome = None

        if name is None:
            name = self._generate_circuit_name()
        self.name = name

        if resource:
            self.counter = Counter(self)

    def __iter__(self):
        """
        Used to iterate commands in quantum circuits.
        """
        return QCircuitIter(self.cmds)

    def allocate(self, qubits, init_state=InitState.ZERO, init_value=0):
        """
        Allocate qubit in quantum circuit.

        Args:
            qubits: number of qubit allocated in circuit.
            init_state: Zero, Plus or Classical state.
            init_value: set the initial value if init_state == Classical

        Returns:
            qreg: the register of quantum.
        """
        self.init_state = init_state
        self.init_value = init_value
        self.qreg = Qureg(circuit=self, size=qubits)
        self.creg = CReg(circuit=self, size=qubits)

        if self.counter:
            self.counter.qubits = qubits

        for index in range(qubits):
            # 建立映射关系
            self.qubit_indices[QuBit(self.qreg, index)] = index
            self.cbit_indices[CBit(self.creg, index)] = index

        return self.qreg

    def get_init_qasm(self):
        """Convert the initialization state of the circuit into qasm instruction."""
        qasm = ""
        if self.init_state == InitState.PLUS:
            qasm += "reset q;\n"
            qasm += "h q;\n"
        elif self.init_state == InitState.CLASSICAL:
            binstr = bin(self.init_value)[2:]
            binarr = [int(b) for b in binstr]
            binarr = binarr[::-1]
            qasm += "reset q;\n"

            for i, val in enumerate(binarr):
                if val == 1:
                    qasm += "x q[" + str(i) + "];\n"
        return qasm

    def set_cmds(self, cmds):
        """
        Set cmds to circuit

        Args:
            cmds: the commands set to circuit
        """
        self.cmds = cmds

    def append_cmd(self, cmd):
        """
        Append command to circuit when apply a quantum gate

        Args:
            cmd: the command append to circuit
        """
        self.cmds.append(cmd)

    @property
    def initstate(self):
        """
        Get circuit init state
        """
        return self.init_state.value

    @property
    def initvalue(self):
        """
        Get circuit init value
        """
        return self.init_value

    def get_density(self):
        """
        Get circuit density
        """
        return self.density

    def forward(self, num):
        """
        Update the cmd_cursor when a bunch of quantum operations have been run.

        Args:
            num: the number of cmds in current bunch
        """
        self.cmd_cursor += num

    @property
    def qubits_len(self):
        """
        Get the number of qubits.
        """
        return len(self.qreg)

    @property
    def gates_len(self):
        """
        Get the number of gates
        """
        return len(self.cmds)

    def set_measure(self, qubit, value):
        """
        Store the measure result for target qubit.

        Args:
            qubit: the index of qubit in qureg
            value: the qubit measure value(0 or 1)
        """
        if qubit >= len(self.qreg) or qubit < 0:
            raise Exception("量子比特下标越界")
        self.creg[qubit].set_value(value)

    def run(self, shots=1):
        """
        Run quantum circuit through the specified backend and shots

        Args:
            shots: run times of the circuit, default: 1

        Returns:
            result: the Result object contain circuit running outcome
        """
        self.backend.send_circuit(self, True)
        result = self.backend.run(shots)
        if result and result.measureSet:
            for m in result.measureSet:
                self.set_measure(m.id, m.value)

        res = Result(self.qubits_len, result)

        return res

    def apply_QFT(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        self.backend.send_circuit(self)
        self.backend.apply_QFT(qubits)

    def apply_Full_QFT(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        self.backend.send_circuit(self)
        self.backend.apply_Full_QFT()

    def draw(self, output=None, line_length=None):
        """
        Draw quantum circuit

        Args:
            output: set the draw device
            line_length: max length to draw quantum circuit, the excess circuit will be truncated
        """
        if output is None:
            output = "text"
        # pylint: disable=C0415
        from qutrunk.visualizations import circuit_drawer

        return circuit_drawer(circuit=self, output=output, line_length=line_length)

    def __str__(self) -> str:
        return str(self.draw(output="text"))

    @property
    def qubits(self) -> List[QuBit]:
        """
        Returns a list of quantum bits
        """
        return self.qreg.qubits

    @property
    def cbits(self) -> List[CBit]:
        """
        Returns a list of cbit
        """
        return self.creg.cbits

    def _generate_circuit_name(self):
        """Generate circuit name"""
        return f"{self.prefix}-{str(random.getrandbits(15))}"

    def __len__(self) -> int:
        """Return number of operations in circuit."""
        return len(self.cmds)

    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        self.backend.send_circuit(self)
        prob = self.backend.get_prob_amp(index)
        return prob

    def get_prob_outcome(self, qubit, outcome):
        """
        Get the probability of a specified qubit being measured in the given outcome (0 or 1)

        Args:
            qubit: the specified qubit to be measured
            outcome: the qubit measure result(0 or 1)

        Returns:
            the probability of target qubit
        """
        self.backend.send_circuit(self)
        prob = self.backend.get_prob_outcome(qubit, outcome)
        return prob

    def get_prob_all_outcome(self, qubits):
        """
        Get outcomeProbs with the probabilities of every outcome of the sub-register contained in qureg

        Args:
            qubits: the sub-register contained in qureg

        Returns:
            An array contains probability of target qubits
        """
        self.backend.send_circuit(self)
        probs = self.backend.get_prob_all_outcome(qubits)
        return probs

    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        self.backend.send_circuit(self)
        state = self.backend.get_all_state()
        return state

    def find_bit(self, bit):
        """Find locations in the circuit.

        Args:
            bit: QuBit or Cbit

        Returns:
            index
        """
        try:
            if isinstance(bit, QuBit):
                return self.qubit_indices[bit]
            elif isinstance(bit, CBit):
                return self.cbit_indices[bit]
            else:
                raise Exception(f"Could not locate bit of unknow type:{type(bit)}")
        except KeyError:
            raise Exception(f"Could not locate provided bit:{bit}")

    def inverse(self):
        """
        Invert this circuit.

        Returns:
            QCircuit: the inverted circuit.
            qreg: the  register of quntum.

        Raises:
            Error: if the circuit cannot be inverted.
        """
        inverse_circuit = QCircuit(backend=self.backend, name=self.name + "_dg")
        inverse_circuit.allocate(qubits=self.qubits_len, init_state=self.init_state, init_value=self.init_value)

        # 反转操作指令，并取门的逆
        cmds = self.cmds
        for cmd in reversed(cmds):
            if isinstance(cmd.gate, MeasureGate):
                raise ValueError("the circuit cannot be inverted.")
            cmd.inverse = True
            inverse_circuit.append_cmd(cmd)

        return inverse_circuit, inverse_circuit.qreg

    @staticmethod
    def from_qasm_file(file):
        """Take in a QASM file and generate a QCircuit object.

        Args:
            file (str): Path to the file for a QASM program
        Return:
            QCircuit: The QCircuit object for the input QASM
        """
        # pylint: disable=C0415，E0611，E1102
        from qutrunk.qasm import Qasm
        from qutrunk.converters import dag_to_circuit
        from qutrunk.converters import ast_to_dag

        qasm = Qasm(file)
        ast = qasm.parse()
        dag = ast_to_dag(ast)
        return dag_to_circuit(dag)

    def expval(self, pauli_prod: Observable):
        """
        Computes the expected value of a product of Pauli operators.

        Args:
            pauliprodlist:
                oper_type (int): Pauli operators
                target (int): indices of the target qubits
        """
        self.backend.send_circuit(self)
        expect = self.backend.get_expec_pauli_prod(pauli_prod.obs_data())
        return expect

    def expval_sum(self, pauli_coeffi: Observable, qubitnum=0):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            pauliprodlist:
                oper_type (int): Pauli operators
                term_coeff (float): The coefficients of each term in the sum of Pauli products
        """
        self.backend.send_circuit(self)
        pauli_type_list, coeffi_list = pauli_coeffi.obs_data()
        if qubitnum != 0:
            if len(coeffi_list) * qubitnum != len(pauli_type_list):
                raise AttributeError("参数错误：参数数量数量的不对")
        expect = self.backend.get_expec_pauli_sum(pauli_type_list, coeffi_list)
        return expect

    def print(self):
        """
        Print quantum circuit in qutrunk form
        """
        print(f"qreg q[{str(len(self.qreg))}]")
        print(f"creg c[{str(len(self.qreg))}]")
        # todo: 增加量子线路初始化转汇编指令处理
        for c in self:
            print(c.quqasm())

    def print_qasm(self, file=None):
        """
        Convert circuit to QASM 2.0, and dump to file/stdout.
        todo: custom gate implemented by qutrunk can't be parsed by third party.
        refer to qulib1.inc and mapping.qutrunk_standard_gate.

        Args:
            file: dump the qasm to file
        """
        f = open(file, "w") if file else sys.stdout

        f.write("OPENQASM 2.0;\n")
        f.write('include "qulib1.inc";\n')
        f.write(f"qreg q[{str(len(self.qreg))}];\n")
        f.write(f"creg c[{str(len(self.qreg))}];\n")
        f.write(self.get_init_qasm())
        for c in self:
            f.write(c.qasm() + ";\n")

        if f is not sys.stdout:
            f.close()

    def show_resource(self):
        """Show resource of the circuit use"""
        if self.counter:
            self.counter.show_verbose()


class QCircuitIter:
    """The iterator for circuit"""

    def __init__(self, cmds):
        self.idx = 0
        self.cmds = cmds

    def __iter__(self):
        return self

    def __next__(self):
        if self.idx < len(self.cmds):
            idx = self.idx
            self.idx += 1
            return self.cmds[idx]
        else:
            raise StopIteration


class Result:
    """Save the result of quantum circuit running"""
    def __init__(self, num_qubits, res):
        self.states = []
        self.values = []
        # 每个量子比特的测量结果
        self.measure_result = [-1] * num_qubits
        if res and res.measureSet:
            for m in res.measureSet:
                self.set_measure(m.id, m.value)

        # 统计所有量子比特组成的比特位字符串出现的次数
        # self.outcome = res.outcomeSet
        if res and res.outcomeSet:
            self.outcome = res.outcomeSet
            for out in self.outcome:
                # 二进制字符串转换成十进制数
                if out.bitstr:
                    self.states.append(int(out.bitstr, base=2))
                    self.values.append(out.count)
        else:
            self.outcome = None

    def set_measure(self, qubit, value):
        """
        Update qubit measure value

        Args:
            qubit: the index of qubit in qureg
            value: the qubit measure value(0 or 1)
        """
        if qubit >= len(self.measure_result) or qubit < 0:
            raise Exception("量子比特下标越界")
        self.measure_result[qubit] = value

    def get_measure(self):
        """Get the measure result"""
        return self.measure_result

    def get_outcome(self):
        """Get the measure result in binary format"""
        out = self.measure_result[::-1]
        bitstr = "0b"
        for o in out:
            bitstr += str(o)
        return bitstr

    def get_counts(self):
        """Get the number of times the measurement results appear"""
        if self.outcome is None:
            return

        res = []
        for out in self.outcome:
            temp = {out.bitstr: out.count}
            res.append(temp)
        return json.dumps(res)



import os
import sys
import pathlib
from setuptools import setup
from setuptools import extension
from distutils.command.build_ext import build_ext
from setuptools import Extension

class CMakeExtension(Extension):
    def __init__(self, name):
        super().__init__(name, sources=[])


simulator = CMakeExtension("qutrunk/sim/local")


ext_modules = [simulator]


class BuildFailed(Exception):
    pass


class ExtBuilder(build_ext):

    def run(self):
        for ext in self.extensions:
            if isinstance(ext, CMakeExtension):
                self.build_cmake(ext)
        super().run()

    def build_cmake(self, ext):
        cwd = pathlib.Path().absolute()

        build_temp = f"{pathlib.Path(self.build_temp)}/{ext.name}"

        os.makedirs(build_temp, exist_ok=True)
        extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
        extdir.mkdir(parents=True, exist_ok=True)

        config = "Debug" if self.debug else "Release"
        cmake_args = [
            "-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=" + str(extdir.parent.absolute()) + "/local",
            "-DCMAKE_BUILD_TYPE=" + config,
            "-DPYTHON_EXECUTABLE=python" + str(sys.version_info.major) + "." + str(sys.version_info.minor)
        ]

        build_args = [
            "--config", config
        ]

        os.chdir(build_temp)
        self.spawn(["cmake", f"{str(cwd)}/{ext.name}"] + cmake_args)
        if not self.dry_run:
            self.spawn(["cmake", "--build", "."] + build_args)
        os.chdir(str(cwd))

def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """
    setup_kwargs.update(
        {"ext_modules": ext_modules, "cmdclass": {"build_ext": ExtBuilder}}
    )


"""
Bell state example.
"""
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, CNOT, Measure


def run_bell_pair(backend=None):
    # allocate
    qc = QCircuit(backend=backend)

    qr = qc.allocate(2)

    # apply gate
    H * qr[0]
    CNOT * (qr[0], qr[1])
    Measure * qr[0]
    Measure * qr[1]

    # print circuit
    qc.print()

    # run circuit
    res = qc.run(shots=100)

    # print result
    print(res.get_measure())
    print(res.get_counts())
    return qc


if __name__ == "__main__":
    circuit = run_bell_pair()

    # 测试draw
    print(circuit.draw())

    # 性能测试
    # import cProfile
    # cProfile.run("run_bell_pair()")



"""Grover's search algorithm."""
import math
import random

from numpy import pi

from qutrunk.circuit import QCircuit, InitState
from qutrunk.circuit.gates import H, X, MCZ
from qutrunk.backends import BackendQuSprout


def apply_oracle(qr, num_qubits, sol_elem):
    for q in range(num_qubits):
        if ((sol_elem >> q) & 1) == 0:
            X * qr[q]

    ctrls = []
    for q in range(num_qubits):
        ctrls.append(qr[q])
    MCZ(len(ctrls) - 1) | (tuple(ctrls))

    for q in range(num_qubits):
        if ((sol_elem >> q) & 1) == 0:
            X * qr[q]


def apply_diffuser(qr, num_qubits):
    for q in range(num_qubits):
        H * qr[q]

    for q in range(num_qubits):
        X * qr[q]

    ctrls = []
    for q in range(num_qubits):
        ctrls.append(qr[q])
    MCZ(len(ctrls) - 1) | (tuple(ctrls))

    for q in range(num_qubits):
        X * qr[q]

    for q in range(num_qubits):
        H * qr[q]


def run_grover(qubits=15, backend=None):
    num_qubits = qubits
    num_elems = 2 ** num_qubits
    num_reps = math.ceil(pi / 4 * math.sqrt(num_elems))
    print("num_qubits:", num_qubits, "num_elems:", num_elems, "num_reps:", num_reps)

    sol_elem = random.randint(0, num_elems - 1)

    circuit = QCircuit(backend=backend, resource=True)

    qureg = circuit.allocate(num_qubits, InitState.PLUS)

    for _ in range(num_reps):
        apply_oracle(qureg, num_qubits, sol_elem)
        apply_diffuser(qureg, num_qubits)
        prob_amp = circuit.get_prob_amp(sol_elem)
        print(f"prob of solution |{sol_elem}> = {prob_amp}")

    # invert circuit
    # circuit, _ = circuit.inverse()
    # 运行线路
    circuit.run()
    # circuit.show_resource()

    return circuit


if __name__ == '__main__':
    # local run
    # circuit = run_grover()

    # qusprout run
    # circuit =run_grover(backend=BackendQuSprout())

    # qusprout mpi run
    # run_grover(backend=BackendQuSprout(ExecType.Mpi))

    # print(circuit.draw(line_length=300))

    # 性能测试
    import cProfile
    cProfile.run("run_grover()")




"""IBM Backend"""

import math

from qutrunk.backends import Backend
from qutrunk.circuit.gates import (
    Measure,
    CNOT,
    H,
    Rx,
    Ry,
    Rz,
)
from .ibm_client import send


# TODO:The IBM quantum chip can only do U1,U2,U3,barriers, and CX / CNOT.
class BackendIBM(Backend):
    """
    The BackendIBM class, which stores the circuit, transforms it to JSON,
    and sends the circuit through the IBM API.
    """
    def __init__(self, token=None, device=None):
        super().__init__()
        # 量子线路
        self.circuit = None
        # token
        self._token = token
        # 默认使用的后端 TODO:后期可选
        self.device = "ibmq_qasm_simulator"
        # 需要分配的量子比特数
        self._allocated_qubits = set()
        # 量子线路的json格式
        self._json = []
        # 测量的id
        self._measured_ids = []

    def send_circuit(self, circuit, final=False):
        """
        Send the quantum circuit to IBM backend

        Args:
            circuit: quantum circuit to send
            final: True if quantum circuit finish, default False
        """
        # 参数：量子比特数
        self._allocated_qubits.add(len(circuit.qreg))

        # 1 遍历circuit
        for ct in circuit:
            self.circuit_to_json(ct)

        # 参数：测量
        for measured_id in self._measured_ids:
            self._json.append({'qubits': [measured_id], 'name': 'measure', 'memory': [measured_id]})
        print("result==", self._json)

        # 2 组装参数
        info = {}
        # 指令集
        info['json'] = self._json

        # 量子比特数
        info["nq"] = len(circuit.qreg)
        # 计算次数
        info['shots'] = 1024
        info['maxCredits'] = 10
        info['backend'] = {'name': self.device}

        # 发送给IBM QE
        result = send(
            info,
            device=self.device,
            token=self._token,
            num_retries=1024,
            verbose=True,
        )
        print("result=", result)
        return result

    def circuit_to_json(self, cmd):
        """Translates the command and in a local variable"""
        gate = cmd.gate

        if gate is Measure:
            self._measured_ids += cmd.targets
        elif gate is CNOT and len(cmd.controls) == 1:
            self._json.append({"qubits": [*cmd.controls, *cmd.targets], "name": "cx"})
        elif gate is H:
            self._json.append({'qubits': cmd.targets, 'name': 'u2', 'params': [0, 3.141592653589793]})
        elif isinstance(gate, (Rx, Ry, Rz)):
            u_name = {'Rx': 'u3', 'Ry': 'u3', 'Rz': 'u1'}
            u_angle = {
                'Rx': [gate.angle, -math.pi / 2, math.pi / 2],
                'Ry': [gate.angle, 0, 0],
                'Rz': [gate.angle],
            }

            # 取出门对应的名称
            gate_name = u_name[str(gate)[0:2]]
            # 取出门对应的角度
            params = u_angle[str(gate)[0:2]]

            self._json.append({'qubits': cmd.targets, 'name': gate_name, 'params': params})

        else:
            raise Exception(
                'This command is not currently supported.\n'
                + 'The IBM quantum chip can only do U1,U2,U3,barriers, and CX / CNOT.'
            )

    def run(self, shots=1):
        pass


"""
Classical register reference object.
"""
import itertools

from .classical_bit import CBit


class CReg:
    """A classical register."""

    prefix = "c"
    bit_type = "CBit"
    instances_counter = itertools.count()

    def __init__(self, circuit=None, name=None, size=None):
        if name is None:
            name = f"{self.prefix}{self.instances_counter}"
        self.name = name

        self.cbits = []
        # 寄存器大小
        self.size = size
        self.circuit = circuit

        for index in range(self.size):
            self.cbits.append(CBit(self, index))

    def append(self, cbit):
        """add a CBit instance"""
        self.cbits.append(cbit)

    def qasm(self):
        """Return OPENQASM string for this register."""
        return f"creg {self.name}[{self.size}];"

    def __getitem__(self, idx):
        """return a Cbit instance.

        Args:
            idx: the index of Cbit.
            
        Returns:
            Cbit instance
        """
        if not isinstance(idx, int):
            raise ValueError("expected integer index into register")
        return self.cbits[idx]

    def __len__(self):
        return len(self.cbits)

    def index(self, cbit):
        """Find the index of the provided cbit within this register."""
        return self.cbits.index(cbit)

    def __setitem__(self, key, value):
        self.cbits[key] = value

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.size})"



from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import CNOT, X
from qutrunk.backends import BackendQuSprout


def run_bernstein_vazirani(backend=None):
    num_qubits = 9
    secret_num = 2 ** 4 + 1
    circuit = QCircuit(backend=backend, resource=True)

    qureg = circuit.allocate(num_qubits)

    X * qureg[0]

    bits = secret_num
    bit = 0
    for qb in range(1, num_qubits):
        bit = int(bits % 2)
        bits = int(bits / 2)
        if bit:
            CNOT * (qureg[0], qureg[qb])

    success_prob = 1.0
    bits = secret_num
    for qb in range(1, num_qubits):
        bit = int(bits % 2)
        bits = int(bits / 2)
        success_prob *= circuit.get_prob_outcome(qb, bit)

    # invert circuit
    # circuit, _ = circuit.inverse()

    print(f"solution reached with probability {success_prob}")
    circuit.print()

    circuit.run()
    circuit.show_resource()

    return circuit


if __name__ == '__main__':
    # local run
    circuit = run_bernstein_vazirani()

    # qusprout run
    # circuit = run_bernstein_vazirani(backend=BackendQuSprout())

    # qusprout mpi run
    # run_bernstein_vazirani(backend=BackendQuSprout(ExecType.Mpi))

    print(circuit.draw())



from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, X, Z, Measure, CNOT, U3


def run_teleport():
    # allocate
    qc = QCircuit()
    qureg = qc.allocate(3)

    # Prepare an initial state
    U3(0.3, 0.2, 0.1) | qureg[0]

    # Prepare a Bell pair
    H * qureg[1]
    CNOT * (qureg[1], qureg[2])

    # Measure in the Bell basis
    CNOT * (qureg[0], qureg[1])
    H * qureg[0]
    Measure * qureg[0]
    Measure * qureg[1]

    # Apply a correction
    Z * qureg[2]
    X * qureg[2]
    Measure * qureg[2]

    # print circuit
    # Printer.print_circuit(qc)

    # run circuit
    res = qc.run(shots=1024)

    print(res.get_counts())
    # return qureg
    return qc


if __name__ == "__main__":
    circuit = run_teleport()
    print(circuit.draw())



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class MCZ(BasicGate):
    """Multi-control Z gate"""
    def __init__(self, ctrl_cnt=1):
        super().__init__()
        self.ctrl_cnt = ctrl_cnt

    def __str__(self):
        return "CZ"

    def __or__(self, qubits):
        """
        Quantum logic gate operation

        Args:
            qubits: the left self.ctrl_cnt qubits are control qubits, the rest right bits are target qubits.

        Example:
            MCZ(2) * (qr[0], qr[1], qr[2]) # qr[0], qr[1] are control qubits, qr[2] is target qubit
        """
        self.qubits = qubits
        controls = [q.index for q in qubits[0:self.ctrl_cnt]]
        targets = [q.index for q in qubits[self.ctrl_cnt:]]
        cmd = Command(self, targets, controls, inverse=self.is_inverse)
        qubits[0].circuit.append_cmd(cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        if self.ctrl_cnt == 1:
            return np.array(
                [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]]
            )


CZ = MCZ(1)



class BasicGate:
    """Base class of all gates. (Don't use it directly but derive from it).

    Args:
        is_inverse: Tells whether to inverse gate or not.
    """

    def __init__(self):
        self.is_inverse = False
        self.name = ""

    def __str__(self):
        """The string description of a quantum logic gate"""
        return ""

    def __or__(self, qubit):
        """Quantum logic gate operation. Overwrite | operator to achieve quantum logic gate operation"""
        pass

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        pass

    def angles(self):
        """Get angle Args"""
        ANGLE_PARAMS = ('rotation', 'theta', 'phi', 'lam', 'gamma')
        return [getattr(self, param) for param in ANGLE_PARAMS if hasattr(self, param)]


class BasicRotateGate(BasicGate):
    """
    Base class for rotation gate
    """

    def __init__(self):
        super().__init__()


class BasciPhaseGate(BasicGate):
    """
    Base class for phase gate
    """

    def __init__(self):
        super().__init__()


class All(BasicGate):
    """
    Meta operator, provides unified operation of multiple qubits

    Args:
        gate: the gate will apply to all qubits
    """

    def __init__(self, gate):
        self.gate = gate

    def __or__(self, qureg):
        """
        Quantum logic gate operation

        Args:
            qureg: the qureg(represent a set of qubit) to apply gate.

        Example:
            All(H) * qureg
            All(Measure) * qureg
        """
        for q in qureg:
            self.gate | q

    def __mul__(self, qureg):
        """
        Args:
            qureg: the qureg(represent a set of qubit) to apply gate.

        Example:
            All(H) * qureg, All(Measure) * qureg
        """
        for q in qureg:
            self.gate * q


def Inv(gate):
    """
    Inverse gate.

    Args:
        gate: the gate will apply inverse operator

    Example:
        Inv(H) * q[0]
    """
    if isinstance(gate, BasicGate):
        gate.is_inverse = True

    return gate



from numpy import pi

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import (
    H, Measure, CNOT, Toffoli, P, R, Rx, Ry, Rz, S, Sdg, T, Tdg, X, Y, Z, MCX, MCZ,
    NOT, Swap, SqrtSwap, SqrtX, All, CP, CY, CZ, CRx, CRy, CRz, Rxx, Ryy, Rzz,
    U1, U2, U3, Barrier, CR, CU, CU1, CU3, CX, iSwap, Y1, X1
)


def run_gates():
    # allocate
    qc = QCircuit()
    qr = qc.allocate(3)

    # apply gate
    H * qr[0]
    CNOT * (qr[0], qr[1])
    NOT * qr[0]
    Toffoli * (qr[0], qr[1], qr[2])
    P(pi / 2) * qr[2]
    R(pi / 2, pi / 2) * qr[0]
    Rx(pi / 2) * qr[1]
    Ry(pi / 2) * qr[1]
    Rz(pi / 2) * qr[1]
    S * qr[0]
    Sdg * qr[0]
    T * qr[0]
    Tdg * qr[0]
    X * qr[2]
    Y * qr[2]
    Z * qr[2]
    X1 * qr[0]
    Y1 * qr[0]
    # Z1 * qr[0]
    Swap * (qr[0], qr[1])
    iSwap(pi / 2) * (qr[0], qr[1])
    SqrtSwap * (qr[0], qr[1])  # 未定义符合
    SqrtX * qr[0]

    CX * (qr[0], qr[1])
    CY * (qr[0], qr[1])
    CZ * (qr[0], qr[1])
    CP(pi / 2) * (qr[0], qr[1])  # 未定义符合
    CR(pi / 2) * (qr[0], qr[1])
    CRx(pi / 2) * (qr[0], qr[1])
    CRy(pi / 2) * (qr[0], qr[1])
    CRz(pi / 2) * (qr[0], qr[1])
    MCX(2) * (qr[0], qr[1], qr[2])
    MCZ(2) * (qr[0], qr[1], qr[2])

    Rxx(pi / 2) * (qr[0], qr[1])
    Ryy(pi / 2) * (qr[0], qr[1])
    Rzz(pi / 2) * (qr[0], qr[1])

    U1(pi / 2) * qr[0]
    U2(pi / 2, pi / 2) * qr[0]
    U3(pi / 2, pi / 2, pi / 2) * qr[0]
    CU(pi / 2, pi / 2, pi / 2, pi / 2) * (qr[0], qr[1])
    CU1(pi / 2) * (qr[1], qr[2])
    CU3(pi / 2, pi / 2, pi / 2) * (qr[0], qr[1])

    Barrier * qr

    All(Measure) * qr

    # print circuit
    qc.print()
    qc.print_qasm()

    qc.run()

    return qc


if __name__ == '__main__':
    circuit = run_gates()
    print(circuit.draw(line_length=3000))
    circuit.print_qasm("allgates-temp.qasm")



from qutrunk.circuit.ops.operation import Pauli, PauliType


class PauliI(Pauli):
    """Apply the PauliI"""
    def __init__(self, *targets):
        super().__init__(list(targets), PauliType.POT_PAULI_I)



import platform

from qutrunk.tools.function_time import timefn
try:
    from qutrunk.sim.local import simulator
except ImportError:  # pragma: no cover
    # windows 
    from qutrunk.sim.local.Release import simulator


class BackendLocalCpp:
    """ Simulator is a compiler engine which simulates a quantum computer using C++-based kernels."""
    @timefn
    def init(self, qubits, state, value, show):
        return simulator.init(qubits, state, value, show)

    @timefn
    def send_circuit(self, circuit, final):
        """
        Send the quantum circuit to local backend

        Args:
            circuit: quantum circuit to send
            final: True if quantum circuit finish, default False, \
                when final==True The backend program will release the computing resources
        """
        start = circuit.cmd_cursor
        cmds = circuit.cmds[start:]
        temp_cmds = []
        for cmd in cmds:
            tempCmd = simulator.Cmd()
            tempCmd.gate = str(cmd.gate)
            tempCmd.targets = cmd.targets
            tempCmd.controls = cmd.controls
            tempCmd.rotation = cmd.rotation
            tempCmd.desc = cmd.qasm()
            tempCmd.inverse = cmd.inverse
            temp_cmds.append(tempCmd)

        simulator.send_circuit(temp_cmds, final)
        # start = circuit.cmd_cursor
        # stop = len(circuit.cmds)
        # for idx in range(start, stop):
        #     cmd = circuit.cmds[idx]
        #     res = simulator.execCmd(gate=str(cmd.gate),
        #                             targets=cmd.targets,
        #                             controls=cmd.controls,
        #                             rotation=cmd.rotation,
        #                             desc=cmd.qasm)

        #     if str(cmd.gate) == "Measure":
        #         qubit = cmd.targets[0]
        #         circuit.qreg[qubit].cbit = res

    @timefn
    def run(self, shots):
        return simulator.run(shots)

    @timefn
    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        return simulator.getProbOfAmp(index)

    @timefn
    def get_prob_outcome(self, qubit, outcome):
        """
        Get the probability of a specified qubit being measured in the given outcome (0 or 1)

        Args:
            qubit: the specified qubit to be measured
            outcome: the qubit measure result(0 or 1)

        Returns:
            the probability of target qubit 
        """
        return simulator.getProbOfOutcome(qubit, outcome)

    @timefn
    def get_prob_all_outcome(self, qubit):
        """
        Get outcomeProbs with the probabilities of every outcome of the sub-register contained in qureg

        Args:
            qubits: the sub-register contained in qureg

        Returns:
            An array contains probability of target qubits
        """
        return simulator.getProbOfAllOutcome(qubit)

    @timefn
    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        return simulator.getAllState()

    @timefn
    def apply_QFT(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        return simulator.apply_QFT(qubits)

    @timefn
    def apply_Full_QFT(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        return simulator.apply_Full_QFT()


from qutrunk.circuit import QCircuit, InitState
from qutrunk.circuit.gates import CNOT, Measure, Toffoli, X, All, MCX


def increment(num_qubits, initvalue):
    """自增运行"""
    circuit = QCircuit()
    qr = circuit.allocate(num_qubits, InitState.CLASSICAL, initvalue)

    # for i in range(num_qubits-1):
    #     ctrl_bits = qr[:num_qubits-1-i]
    #     C(NOT, num_qubits-1-i) | (*ctrl_bits, qr[num_qubits-1-i])
    # X | qr[0]

    # 等同于下面的代码
    MCX(3) * (qr[0], qr[1], qr[2], qr[3])
    Toffoli * (qr[0], qr[1], qr[2])
    CNOT * (qr[0], qr[1])
    X * qr[0]

    All(Measure) * qr
    res = circuit.run()
    print(res.get_outcome())

    return circuit


def decrement(num_qubits, initvalue):
    """自减运行"""
    circuit = QCircuit()
    qr = circuit.allocate(num_qubits, InitState.CLASSICAL, initvalue)

    # X | qr[0]
    # for i in range(num_qubits-1):
    #     ctrl_bits = qr[:1+i]
    #     C(NOT, 1+i) | (*ctrl_bits, qr[1+i])

    # 等同于下面的代码
    X * qr[0]
    CNOT * (qr[0], qr[1])
    Toffoli * (qr[0], qr[1], qr[2])
    MCX(3) * (qr[0], qr[1], qr[2], qr[3])

    All(Measure) * qr
    res = circuit.run()
    print(res.get_outcome())
    return circuit


if __name__ == "__main__":
    circuit = increment(4, 0)

    # circuit = decrement(4, 0b0001)

    print(circuit.draw())



import itertools

from .qubit import QuBit


class Qureg:
    """
    Qubit register, maintains a set of qubits.

    Arg:
        name: The name of the register.
        qubits: a list which store Qubit instance
    """
    prefix = "q"
    bit_type = "QuBit"
    instances_counter = itertools.count()

    def __init__(self, circuit=None, name=None, size=None):
        if name is None:
            name = f"{self.prefix}{self.instances_counter}"
        self.name = name

        self.qubits = []
        # 寄存器大小
        self.size = size
        self.circuit = circuit

        for index in range(self.size):
            self.qubits.append(QuBit(self, index))

    def __getitem__(self, idx):
        """return a Qubit instance.
        Arg:
            idx: the index of Qubit.
        Returns:
            Qubit instance
        """
        if not isinstance(idx, int):
            raise ValueError("expected integer index into register")
        return self.qubits[idx]

    def __len__(self):
        return len(self.qubits)

    def qasm(self):
        """Return OPENQASM string for this register."""
        return f"qreg q[{len(self)}];"

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.size})"

    def index(self, qubit):
        """Find the index of the provided qubit within this register."""
        return self.qubits.index(qubit)








from qutrunk.backends.backend import Backend

from qutrunk.tools.read_qubox import get_qulocalbox_setting

try:
    from qutrunk.sim.local.local_cpp import BackendLocalCpp as BackendLocalImpl
except ImportError:
    from qutrunk.sim.local.local_python import BackendLocalPython as BackendLocalImpl

from enum import Enum

class BackendLocalType(Enum):
    """Backend Local Type"""

    CPP = 0
    PYTHON = 1
    UNKNOWN = 2

class BackendLocal(Backend):
    """
    The local backend uses the simulator to run the quantum circuit, qutrunk provide two types simulator.
    C++ simulator is preferred. If C++ simulator is not available, python simulator is used instead
    """
    def __init__(self):
        super().__init__()
        self.circuit = None
        self._local_impl = BackendLocalImpl()
        box_config = get_qulocalbox_setting()
        self._show_quantum_gate = box_config.get('show_quantum_gate')

    def send_circuit(self, circuit, final=False):
        """
        Send the quantum circuit to local backend

        Args:
            circuit: quantum circuit to send
            final: True if quantum circuit finish, default False, \
                when final==True The backend program will release the computing resources 
        """
        start = circuit.cmd_cursor
        stop = len(circuit.cmds)

        if start == 0:
            res, elapsed = self._local_impl.init(
                len(circuit.qreg), 
                circuit.initstate, 
                circuit.initvalue,
                self._show_quantum_gate)
            if self.circuit.counter:
                self.circuit.counter.acc_run_time(elapsed)

        res, elapsed = self._local_impl.send_circuit(circuit, final)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)

        circuit.forward(stop - start)

        return res

    def run(self, shots=1):
        """
        Run quantum circuit

        Args:
            shots: circuit run times, for sampling, default: 1
        
        Returns:
            result: the Result object contain circuit running outcome
        """
        res, elapsed = self._local_impl.run(shots)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
            self.circuit.counter.finish()
        return res

    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        res, elapsed = self._local_impl.get_prob_amp(index)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_outcome(self, qubit, outcome):
        """
        Get the probability of a specified qubit being measured in the given outcome (0 or 1)

        Args:
            qubit: the specified qubit to be measured
            outcome: the qubit measure result(0 or 1)

        Returns:
            the probability of target qubit
        """
        res, elapsed = self._local_impl.get_prob_outcome(qubit, outcome)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_prob_all_outcome(self, qubits):
        """
        Get outcomeProbs with the probabilities of every outcome of the sub-register contained in qureg

        Args:
            qubits: the sub-register contained in qureg

        Returns:
            An array contains probability of target qubits
        """
        res, elapsed = self._local_impl.get_prob_all_outcome(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        res, elapsed = self._local_impl.get_all_state()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def apply_QFT(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        res, elapsed = self._local_impl.apply_qft(qubits)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def apply_Full_QFT(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        res, elapsed = self._local_impl.apply_full_qft()
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_expec_pauli_prod(self, pauli_prod_list):
        """
        Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: a list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            the expected value of a product of Pauli operators.
        """
        res, elapsed = self._local_impl.get_expec_pauli_prod(pauli_prod_list)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res
    
    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: a list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: the coefficients of each term in the sum of Pauli products.

        Returns:
            the expected value of a sum of products of Pauli operators. 
        """
        res, elapsed = self._local_impl.get_expec_pauli_sum(oper_type_list, term_coeff_list)
        if self.circuit.counter:
            self.circuit.counter.acc_run_time(elapsed)
        return res

    def get_local_type(self):
        """
        return which local is being used
        """
        if type(self._local_impl) == BackendLocalCpp:
            return BackendLocalType.CPP
        elif type(self._local_impl) == BackendLocalPython:
            return BackendLocalType.PYTHON
        else:
            return BackendLocalType.UNKNOWN


from qutrunk.circuit.qubit import QuBit
from qutrunk.circuit.command import Command
from .basicgate import BasicGate


class BarrierGate(BasicGate):
    """Barrier Gate"""

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Barrier"

    def __or__(self, qubits):
        """
        Quantum logic gate operation

        Args:
            qubits: the quantum bit to aplly Barrier gate.

        Example:
            Barrier * (qr[0], qr[1])
        """
        templist = [qubits] if isinstance(qubits, QuBit) else qubits
        targets = [q.index for q in templist]
        cmd = Command(self, targets)
        templist[0].circuit.append_cmd(cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubits)

    @property
    def label(self):
        """Return gate label"""
        return "Barrier"


Barrier = BarrierGate()



# TODO:模块内统一包引用
import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicRotateGate


class CP(BasicRotateGate):
    """Control Phase Gate"""
    def __init__(self, angle):
        super().__init__()
        self.rotation = angle

    def __str__(self):
        return "CP"

    def __or__(self, qubits):
        """
        Quantum logic gate operation

        Args:
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.

        Example:
            CP * (qr[0], qr[1])
        """
        if len(qubits) != 2:
            raise AttributeError("参数错误：应该传入1个控制位，1个目标位")
        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(self, targets, controls, rotation=[self.rotation], inverse=self.is_inverse)
        qubits[0].circuit.append_cmd(cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        eith = np.exp(1j * float(self.rotation))
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, eith, 0], [0, 0, 0, 1]])



import uuid
from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TMultiplexedProtocol

from qutrunk.sim.qusprout.qusprout import QuSproutServer
from qutrunk.sim.qusprout.work import WorkServer

from qutrunk.sim.qusprout.qusproutdata import ttypes as qusproutdata

from qutrunk.tools.function_time import timefn

class WorkApiServer:
    def __init__(self, ip='localhost', port=9090):
        socket = TSocket.TSocket(ip, port)
        self._transport = TTransport.TBufferedTransport(socket)
        protocol = TBinaryProtocol.TBinaryProtocol(self._transport)
        quest = TMultiplexedProtocol.TMultiplexedProtocol(protocol, "WorkServer")
        self._client = WorkServer.Client(quest)
        self._taskid = uuid.uuid4().hex
        self._transport.open()

    def close(self):
        self._transport.close()

    @timefn
    def init(self, qubits, state, value, density, exectype=0):
        req = qusproutdata.InitQubitsReq(self._taskid, qubits, state, value, density, exectype)
        return self._client.initQubits(req)

    @timefn
    def send_circuit(self, circuit, final):
        req = qusproutdata.SendCircuitCmdReq(self._taskid, circuit, final)
        res = self._client.sendCircuitCmd(req)
        return res

    @timefn
    def run(self, shots):
        req = qusproutdata.RunCircuitReq(self._taskid, shots)
        res = self._client.run(req)
        return res.result

    @timefn
    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        req = qusproutdata.GetProbAmpReq(self._taskid, index)
        res = self._client.getProbAmp(req)
        return res.amp

    @timefn
    def get_prob_outcome(self, qubit, outcome):
        """
        Get the probability of a specified qubit being measured in the given outcome (0 or 1)

        Args:
            qubit: the specified qubit to be measured
            outcome: the qubit measure result(0 or 1)

        Returns:
            the probability of target qubit 
        """
        req = qusproutdata.GetProbOfOutcomeReq(self._taskid, qubit, outcome)
        res = self._client.getProbOfOutcome(req)
        return res.pro_outcome

    @timefn
    def get_prob_all_outcome(self, qubits):
        """
        Get outcomeProbs with the probabilities of every outcome of the sub-register contained in qureg

        Args:
            qubits: the sub-register contained in qureg

        Returns:
            An array contains probability of target qubits
        """
        req = qusproutdata.GetProbOfAllOutcomReq(self._taskid, qubits)
        res = self._client.getProbOfAllOutcome(req)
        return res.pro_outcomes

    @timefn
    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        req = qusproutdata.GetAllStateReq(self._taskid)
        res = self._client.getAllState(req)
        return res.all_state

    @timefn
    def cancel_Cmd(self):
        req = qusproutdata.CancelCmdReq(self._taskid)
        return self._client.cancelCmd(req)

    @timefn
    def apply_QFT(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        req = qusproutdata.ApplyQFTReq(self._taskid, qubits)
        res = self._client.applyQFT(req)

    @timefn
    def apply_Full_QFT(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        req = qusproutdata.ApplyFullQFTReq(self._taskid)
        res = self._client.applyFullQFT(req)

    @timefn
    def get_expec_pauli_prod(self, paulilist):
        """
        Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: a list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            the expected value of a product of Pauli operators.
        """
        req = qusproutdata.GetExpecPauliProdReq(self._taskid, paulilist)
        res = self._client.getExpecPauliProd(req)
        return res
    
    @timefn
    def get_expec_pauli_sum(self, paulilist):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: a list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: the coefficients of each term in the sum of Pauli products.

        Returns:
            the expected value of a sum of products of Pauli operators. 
        """
        req = qusproutdata.GetExpecPauliSumReq(self._taskid, paulilist)
        res = self._client.getExpecPauliSum(req)
        return res
    
class QuSproutApiServer:
    def __init__(self, ip='localhost', port=9090):
        socket = TSocket.TSocket(ip, port)
        self._transport = TTransport.TBufferedTransport(socket)
        protocol = TBinaryProtocol.TBinaryProtocol(self._transport)
        quest = TMultiplexedProtocol.TMultiplexedProtocol(protocol, "QuSproutServer")
        self._client = QuSproutServer.Client(quest)
        self._taskid = uuid.uuid4().hex
        self._transport.open()

    def close(self):
        self._transport.close()

    @timefn
    def init(self, qubits, state, value, density, exectype=0):
        req = qusproutdata.InitQubitsReq(self._taskid, qubits, state, value, density, exectype)
        return self._client.initQubits(req)

    @timefn
    def send_circuit(self, circuit, final):
        req = qusproutdata.SendCircuitCmdReq(self._taskid, circuit, final)
        res = self._client.sendCircuitCmd(req)
        return res

    @timefn
    def run(self, shots):
        req = qusproutdata.RunCircuitReq(self._taskid, shots)
        res = self._client.run(req)
        return res.result

    @timefn
    def get_prob_amp(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        req = qusproutdata.GetProbAmpReq(self._taskid, index)
        res = self._client.getProbAmp(req)
        return res.amp

    @timefn
    def get_prob_outcome(self, qubit, outcome):
        """
        Get the probability of a specified qubit being measured in the given outcome (0 or 1)

        Args:
            qubit: the specified qubit to be measured
            outcome: the qubit measure result(0 or 1)

        Returns:
            the probability of target qubit 
        """
        req = qusproutdata.GetProbOfOutcomeReq(self._taskid, qubit, outcome)
        res = self._client.getProbOfOutcome(req)
        return res.pro_outcome

    @timefn
    def get_prob_all_outcome(self, qubits):
        """
        Get outcomeProbs with the probabilities of every outcome of the sub-register contained in qureg

        Args:
            qubits: the sub-register contained in qureg

        Returns:
            An array contains probability of target qubits
        """
        req = qusproutdata.GetProbOfAllOutcomReq(self._taskid, qubits)
        res = self._client.getProbOfAllOutcome(req)
        return res.pro_outcomes

    @timefn
    def get_all_state(self):
        """
        Get the current state vector of probability amplitudes for a set of qubits
        """
        req = qusproutdata.GetAllStateReq(self._taskid)
        res = self._client.getAllState(req)
        return res.all_state

    @timefn
    def cancel_Cmd(self):
        req = qusproutdata.CancelCmdReq(self._taskid)
        return self._client.cancelCmd(req)

    @timefn
    def apply_QFT(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        req = qusproutdata.ApplyQFTReq(self._taskid, qubits)
        res = self._client.applyQFT(req)

    @timefn
    def apply_Full_QFT(self):
        """
        Applies the quantum Fourier transform (QFT) to the entirety of qureg
        """
        req = qusproutdata.ApplyFullQFTReq(self._taskid)
        res = self._client.applyFullQFT(req)

    @timefn
    def get_expec_pauli_prod(self, pauli_prod_list):
        """
        Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: a list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            the expected value of a product of Pauli operators.
        """
        req = qusproutdata.GetExpecPauliProdReq(self._taskid, pauli_prod_list)
        res = self._client.getExpecPauliProd(req)
        return res.expect
    
    @timefn
    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: a list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: the coefficients of each term in the sum of Pauli products.

        Returns:
            the expected value of a sum of products of Pauli operators. 
        """
        req = qusproutdata.GetExpecPauliSumReq(self._taskid, oper_type_list, term_coeff_list)
        res = self._client.getExpecPauliSum(req)
        return res.expect
        


"""Example of a simple quantum random number generator."""

# https://blog.csdn.net/longji/article/details/85547108?ops_request_misc=%257B%2522request%255Fid%2522%253A%2522164756639816780261973236%2522%252C%2522scm%2522%253A%252220140713.130102334..%2522%257D&request_id=164756639816780261973236&biz_id=0&utm_medium=distribute.pc_search_result.none-task-blog-2~all~sobaiduend~default-1-85547108.142^v2^pc_search_insert_es_download,143^v4^control&utm_term=projectq&spm=1018.2226.3001.4187

from qutrunk.backends import BackendQuSprout
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, All, X, CNOT


def Deutsch(backend=None):
    # allocate
    qc = QCircuit(backend=backend)
    qureg = qc.allocate(2)
    X * qureg[1]
    All(H) * qureg
    CNOT * (qureg[0], qureg[1])
    H * qureg[0]
    All(Measure) * qureg
    qc.print()
    result = qc.run()
    # TODO:??
    print("执行Deutsch算法，测量得到的结果是：", result.get_measure()[0])
    return qc


if __name__ == "__main__":
    # local run
    circuit = Deutsch()

    # qusprout run
    # circuit = Deutsch(backend=BackendQuSprout())
    print(circuit.draw())



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class U3(BasicGate):
    """
    U3 gate
    """

    def __init__(self, theta, phi, lam):
        """
        Args:
            theta: U3 gate parameter1
            phi: U3 gate parameter2
            lam: U3 gate parameter3
        """
        super().__init__()
        self.theta = theta
        self.phi = phi
        self.lam = lam

    def __str__(self):
        return "U3"

    def __or__(self, qubit):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to apply U3 gate.

        Example:
            U3(pi/2,pi/2,pi/2) * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.theta, self.phi, self.lam], inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        theta = self.theta
        phi = self.phi
        lam = self.lam
        cos = np.cos(theta / 2)
        sin = np.sin(theta / 2)
        return np.array(
            [
                [cos, -np.exp(1j * lam) * sin],
                [np.exp(1j * phi) * sin, np.exp(1j * (phi + lam)) * cos],
            ]
        )


class U(U3):
    pass



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicRotateGate


class CR(BasicRotateGate):
    """Control rotation gate"""
    def __init__(self, alpha):
        """
        Args:
            alpha: rotation angle
        """
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "CR"

    def __or__(self, qubits):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to aplly CR gate

        Example:
            CR(pi/2) * (qr[0], qr[1])
        """
        if len(qubits) != 2:
            raise AttributeError("Parameter Error: qubits should be two")
        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(self, targets, controls, inverse=self.is_inverse, rotation=[self.rotation])
        qubits[0].circuit.append_cmd(cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        half_alpha = float(self.rotation)
        return np.matrix([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, np.exp(1j * half_alpha)]])



import numpy as np
import cmath
import math

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import (
    H, Measure, CNOT, Toffoli, P, R, Rx, Ry, Rz, S, Sdg, T, Tdg, 
    X, Y, Z, NOT, Swap, SqrtSwap, SqrtX, All, Rxx, Ryy, Rzz, U1, U2, U3,
    CP, CRx, CRy, CRz, CX, CY, CZ
)


def test_gate_matrix():
    angle = np.pi/2
    cos = math.cos(angle / 2)
    sin = math.sin(angle / 2)
    exp_m = np.exp(-1j * angle)
    exp_p = np.exp(1j * angle)
    assert np.array_equal(H.matrix, 1.0 / cmath.sqrt(2.0) * np.matrix([[1, 1], [1, -1]]))
    assert np.array_equal(P(angle).matrix, np.matrix([[cmath.exp(1j * angle), 0], [0, cmath.exp(1j * angle)]]))
    assert np.array_equal(R(angle, angle).matrix, np.matrix([[cos, -1j * exp_m * sin], [-1j * exp_p * sin, cos]]))
    assert np.array_equal(Rx(angle).matrix, np.matrix(
            [
                [math.cos(0.5 * angle), -1j * math.sin(0.5 * angle)],
                [-1j * math.sin(0.5 * angle), math.cos(0.5 * angle)],
            ]
        ))
    assert np.array_equal(Ry(angle).matrix, np.matrix(
            [
                [math.cos(0.5 * angle), -math.sin(0.5 * angle)],
                [math.sin(0.5 * angle), math.cos(0.5 * angle)],
            ]
        ))
    assert np.array_equal(Rz(angle).matrix, np.matrix(
            [
                [cmath.exp(-0.5 * 1j * angle), 0],
                [0, cmath.exp(0.5 * 1j * angle)],
            ]
        ))
    assert np.array_equal(S.matrix, np.matrix([[1, 0], [0, 1j]]))
    assert np.array_equal(Sdg.matrix, np.matrix([[1, 0], [0, 1j]]).getH())
    assert np.array_equal(T.matrix, np.matrix([[1, 0], [0, cmath.exp(1j * cmath.pi / 4)]]))
    assert np.array_equal(Tdg.matrix, np.matrix([[1, 0], [0, cmath.exp(1j * cmath.pi / 4)]]).getH())
    assert np.array_equal(X.matrix, np.matrix([[0, 1], [1, 0]]))
    assert np.array_equal(Y.matrix, np.matrix([[0, -1j], [1j, 0]]))
    assert np.array_equal(Z.matrix, np.matrix([[1, 0], [0, -1]]))
    assert np.array_equal(NOT.matrix, X.matrix)
    assert np.array_equal(Swap.matrix, np.matrix([[1, 0, 0, 0],
                          [0, 0, 1, 0],
                          [0, 1, 0, 0],
                          [0, 0, 0, 1]]))
    assert np.array_equal(SqrtSwap.matrix, np.matrix(
            [
                [1, 0, 0, 0],
                [0, 0.5 + 0.5j, 0.5 - 0.5j, 0],
                [0, 0.5 - 0.5j, 0.5 + 0.5j, 0],
                [0, 0, 0, 1],
            ]
        ))
    assert np.array_equal(SqrtX.matrix, 0.5 * np.matrix([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]]))
    assert np.array_equal(Rxx(angle).matrix, np.matrix(
            [
                [cmath.cos(0.5 * angle), 0, 0, -1j * cmath.sin(0.5 * angle)],
                [0, cmath.cos(0.5 * angle), -1j * cmath.sin(0.5 * angle), 0],
                [0, -1j * cmath.sin(0.5 * angle), cmath.cos(0.5 * angle), 0],
                [-1j * cmath.sin(0.5 * angle), 0, 0, cmath.cos(0.5 * angle)],
            ]
        ))
    assert np.array_equal(Ryy(angle).matrix, np.matrix(
            [
                [cmath.cos(0.5 * angle), 0, 0, 1j * cmath.sin(0.5 * angle)],
                [0, cmath.cos(0.5 * angle), -1j * cmath.sin(0.5 * angle), 0],
                [0, -1j * cmath.sin(0.5 * angle), cmath.cos(0.5 * angle), 0],
                [1j * cmath.sin(0.5 * angle), 0, 0, cmath.cos(0.5 * angle)],
            ]
        ))
    assert np.array_equal(Rzz(angle).matrix, np.matrix(
            [
                [cmath.exp(-0.5 * 1j * angle), 0, 0, 0],
                [0, cmath.exp(0.5 * 1j * angle), 0, 0],
                [0, 0, cmath.exp(0.5 * 1j * angle), 0],
                [0, 0, 0, cmath.exp(-0.5 * 1j * angle)],
            ]
        ))
    assert np.array_equal(U1(angle).matrix, np.matrix([[1, 0], [0, np.exp(1j * angle)]]))
    isqrt2 = 1 / np.sqrt(2)
    assert np.array_equal(U2(angle, angle).matrix, np.matrix(
            [
                [isqrt2, -np.exp(1j * angle) * isqrt2],
                [np.exp(1j * angle) * isqrt2, np.exp(1j * (angle + angle)) * isqrt2],
            ]
        ))
    assert np.array_equal(U3(angle, angle, angle).matrix, np.array(
            [
                [np.cos(angle/2), -np.exp(1j * angle) * np.sin(angle/2)],
                [np.exp(1j * angle) * np.sin(angle/2), np.exp(1j * (angle + angle)) * np.cos(angle/2)],
            ]
        ))
    eith = np.exp(1j * float(angle))
    assert np.array_equal(CP(angle).matrix, np.array(
        np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, eith, 0], [0, 0, 0, 1]])
    ))
    half_theta = angle / 2
    cos = np.cos(half_theta)
    sin = np.sin(half_theta)
    isin = 1j * np.sin(half_theta)
    assert np.array_equal(CRx(angle).matrix, np.array(
        np.array([[cos, 0, -isin, 0], [0, 1, 0, 0], [-isin, 0, cos, 0], [0, 0, 0, 1]])
    ))
    assert np.array_equal(CRy(angle).matrix, np.array(
        np.array([[cos, 0, -sin, 0], [0, 1, 0, 0], [sin, 0, cos, 0], [0, 0, 0, 1]])
    ))
    arg = 1j * half_theta
    assert np.array_equal(CRz(angle).matrix, np.array(
        np.array([[np.exp(-arg), 0, 0, 0], [0, 1, 0, 0], [0, 0, np.exp(arg), 0], [0, 0, 0, 1]])
    ))
    assert np.array_equal(CX.matrix, np.array(
        np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
    ))
    assert np.array_equal(CY.matrix, np.array(
        np.array([[0, 0, -1j, 0], [0, 1, 0, 0], [1j, 0, 0, 0], [0, 0, 0, 1]])
    ))
    assert np.array_equal(CZ.matrix, np.array(
        np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
    ))


import yaml
import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_qubox_setting():
    """Get Qubox configure info"""
    with open(BASE_DIR + "/config/qubox.yaml") as f:
        yaml_content = yaml.load(f, Loader=yaml.FullLoader)
        return yaml_content


def get_qulocalbox_setting():
    """Get local backend configure info"""
    with open(BASE_DIR + "/config/qulocal.yaml") as f:
        yaml_content = yaml.load(f, Loader=yaml.FullLoader)
        return yaml_content



from .bit import Bit


class QuBit(Bit):
    """
    a quantum bit

    Attributes:
        circuit: A quantum circuit.
        index: The index of the bit in its containing register.
    """

    def __init__(self, qreg=None, index=None):
        """Creates a quantum bit."""
        super().__init__(qreg, index)
        if qreg is not None:
            self.circuit = qreg.circuit



from .bit import Bit


class CBit(Bit):
    """Implement a classical bit."""

    def __init__(self, creg=None, index=None, value=None):
        """Creates a classical bit."""
        super().__init__(creg, index)
        # value保存测量的值
        self.value = value
        if creg is not None:
            self.circuit = creg.circuit

    def set_value(self, value):
        """Update cbit value when the corresponding qubit measure done"""
        self.value = value



from qutrunk.circuit import QCircuit, Counter
from qutrunk.circuit.gates import H, Measure, All
from qutrunk.backends import BackendQuSprout


def run_random_byte(backend=None):
    # allocate
    qc = QCircuit(backend)

    qureg = qc.allocate(8)
    # 将统计信息归集到Counter类
    ct = Counter(qc)

    All(H) * qureg
    All(Measure) * qureg

    # print circuit
    qc.print()

    # run circuit
    res = qc.run()

    # print([int(q) for q in qureg])
    print(res.get_measure())
    ct.show_verbose()
    return qc


if __name__ == "__main__":
    # local run
    circuit = run_random_byte()

    # qusprout run
    # circuit = run_random_byte(backend=BackendQuSprout())

    # qusprout mpi run
    # circuit = run_random_byte(backend=BackendQuSprout(ExecType.Mpi))

    print(circuit.draw())



"""
GHZ state example.
"""
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import CX, Measure, H, Barrier, All
from qutrunk.backends import BackendQuSprout


def run_ghz(qubits=5, backend=None):
    # Allocate
    qc = QCircuit(name="ghz", backend=backend)
    qr = qc.allocate(qubits)

    # Create a GHZ state
    H * qr[0]
    for i in range(qubits - 1):
        CX * (qr[i], qr[i + 1])

    Barrier * qr
    # Measure all of the qubits
    All(Measure) * qr

    # Run 1024 times
    res = qc.run(shots=1024)
    print(res.get_counts())
    return qc


if __name__ == "__main__":
    # circuit = run_ghz()

    # qusprout run
    # circuit = run_ghz(backend=BackendQuSprout())

    # print(circuit.draw())

    # 性能测试
    import cProfile
    cProfile.run("run_ghz()")




from .barrier import Barrier, BarrierGate
from .basicgate import All, Inv, BasicGate
from .cp import CP
from .cr import CR
from .crx import CRx
from .cry import CRy
from .crz import CRz
from .cu import CU
from .cu1 import CU1
from .cu3 import CU3
from .cx import CX, CNOT, Toffoli, MCX
from .cy import CY, CYGate
from .cz import CZ, MCZ
from .h import H, HGate
from .iswap import iSwap
from .measure import Measure, MeasureGate
from .p import P
from .r import R
from .rx import Rx
from .rxx import Rxx
from .ry import Ry
from .ryy import Ryy
from .rz import Rz
from .rzz import Rzz
from .s import S, SGate
from .sdg import Sdg, SdgGate
from .sqrtswap import SqrtSwap, SqrtSwapGate
from .sqrtx import SqrtX, SqrtXGate
from .swap import Swap, SwapGate
from .t import T, TGate
from .tdg import Tdg, TdgGate
from .u1 import U1
from .u2 import U2
from .u3 import U3, U
from .x import X, NOT, XGate
from .x1 import X1, X1Gate
from .y import Y, YGate
from .y1 import Y1, Y1Gate
from .z import Z, ZGate
from .z1 import Z1, Z1Gate

CTRL_CNT_PARAM = "ctrl_cnt"




from abc import ABCMeta, abstractmethod


class Backend(metaclass=ABCMeta):
    """
    Basic simulator: All simulators are derived from this class.
    """
    @abstractmethod
    def send_circuit(self, circuit, final=False):
        """
        send the quantum circuit to backend
        """
        pass



"""
Quantum Fourier Transform examples.
"""
import numpy as np

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, All, P


def run_QFT():
    circuit = QCircuit()
    qureg = circuit.allocate(3)

    All(H) * qureg
    circuit.apply_QFT([q.index for q in qureg])
    print(circuit.get_all_state())

    circuit.run()


def run_Full_QFT():
    circuit = QCircuit()
    qureg = circuit.allocate(3)

    All(H) * qureg
    circuit.apply_Full_QFT()
    print(circuit.get_all_state())

    circuit.run()


def qft_single_wave():
    num_qubits = 4
    circuit = QCircuit()
    qreg = circuit.allocate(num_qubits)
    All(H) * qreg
    P(np.pi / 4) * qreg[0]
    P(np.pi / 2) * qreg[1]
    P(np.pi) | qreg[2]
    circuit.apply_Full_QFT()
    print(circuit.get_all_state())
    circuit.run()

    return circuit


if __name__ == "__main__":
    # run_QFT()
    # run_Full_QFT()
    circuit = qft_single_wave()
    print(circuit.draw())



import sys
sys.path.append("D:/code/qike/qubox/src") 
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, All, X1, Y1, Z1
from qutrunk.backends import BackendQuSprout, ExecType

def testx1(backend=None):
    # allocate
    qc = QCircuit(backend=backend)
    qureg = qc.allocate(1)
    H | qureg[0]
    X1 | qureg[0]

    #qc, qureg = qc.inverse()

    All(Measure) | qureg
    qc.print()
    result = qc.run()

    print("执行X1，测量得到的结果是：", result.get_measure()[0])
    return qc

def testy1(backend=None):
    # allocate
    qc = QCircuit(backend=backend)
    qureg = qc.allocate(1)
    H | qureg[0]
    Y1 | qureg[0]

    qc, qureg = qc.inverse()

    All(Measure) | qureg
    qc.print()
    result = qc.run()

    print("执行Y1，测量得到的结果是：", result.get_measure()[0])
    return qc

def testz1(backend=None):
    # allocate
    qc = QCircuit(backend=backend)
    qureg = qc.allocate(1)
    H | qureg[0]
    Z1 | qureg[0]

    qc, qureg = qc.inverse()

    All(Measure) | qureg
    qc.print()
    result = qc.run()

    print("执行Z1，测量得到的结果是：", result.get_measure()[0])
    return qc

if __name__ == "__main__":
    qc = testx1()
    print(qc.draw())

    qcy = testy1()
    print(qcy.draw())

    qcz = testz1()
    print(qcz.draw())





"""
The entry for drawing quantum circuit diagram
"""
from .utils import _get_instructions
from .text import TextDrawing


def circuit_drawer(circuit, output, line_length):
    """Select the corresponding circuit drawer according to the output parameter"""
    if output == "text":
        return __text_circuit_drawer(circuit=circuit, line_length=line_length)


def __text_circuit_drawer(circuit, plot_barriers=True, vertical_compression="high", line_length=None):
    """Draw quantum circuit by ascii art"""
    # 1 获取量子比特位数、经典比特位数、DAG的节点
    qubits, cbits, nodes = _get_instructions(circuit)
    # 2 绘制量子电路
    text_drawing = TextDrawing(qubits=qubits, cbits=cbits, nodes=nodes, circuit=circuit)
    # 是否输出barriers，默认为True，即输出
    text_drawing.plot_barriers = plot_barriers
    # 画图时，线的最大长度
    text_drawing.line_length = line_length
    # text方式的垂直压缩方式，默认是：medium
    text_drawing.vertical_compression = vertical_compression

    # 3 返回绘制结果
    return text_drawing



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class YGate(BasicGate):
    """Apply the single-qubit Pauli-Y (also known as the Y or sigma-Y) gate"""

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Y"

    def __or__(self, qubit):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to apply Y gate.

        Example:
            Y * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[0, -1j], [1j, 0]])


Y = YGate()



from qutrunk.circuit.gates import (
    MCX, CYGate, MCZ, SwapGate, HGate, SGate, SdgGate, SqrtXGate,
    TGate, TdgGate, P, U1, U2, U3, U, CU, CU1, CU3,
    XGate, YGate, ZGate, Rx, Ry, Rz, Rxx, Rzz, CP, CRx, CRy, CRz, R, Ryy, SqrtSwapGate, iSwap, CR, X1Gate,
    Y1Gate, Z1Gate
)

# qasm的量子门映射到qutrunk的量子门
qutrunk_standard_gate = {
    # standard gate from qasm
    "u1": U1,
    "u2": U2,
    "u3": U3,
    "u": U,
    "p": P,
    "x": XGate,
    "y": YGate,
    "z": ZGate,
    "t": TGate,
    "tdg": TdgGate,
    "s": SGate,
    "sdg": SdgGate,
    "sx": SqrtXGate,
    # "sxdg": SXdgGate, # not implemented, decomposition in qelib1.inc
    "swap": SwapGate,
    "rx": Rx,
    "rxx": Rxx,
    "ry": Ry,
    "rz": Rz,
    "rzz": Rzz,
    # "id": IGate,
    "h": HGate,
    "cx": MCX,
    "cy": CYGate,
    "cz": MCZ,
    # "ch": CHGate, # not implemented, decomposition in qelib1.inc
    "crx": CRx,
    "cry": CRy,
    "crz": CRz,
    # "csx": CSXGate, # not implemented, decomposition in qelib1.inc
    "cu1": CU1,
    "cp": CP,
    "cu": CU,
    "cu3": CU3,
    # "ccx": MCX(2), # qutrunk implement by MCX
    # "cswap": CSwapGate, # not implemented, decomposition in qelib1.inc
    # "delay": Delay,

    # qutrunk only gate
    "r": R,
    "ryy": Ryy,
    "sqrtswap": SqrtSwapGate,
    "iswap": iSwap,
    "cr": CR,
    "x1": X1Gate,
    "y1": Y1Gate,
    "z1": Z1Gate,
    "mcx": MCX,
    "mcz": MCZ
}



import math

import pytest
from numpy import pi

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import CP
from qutrunk.test.gate.global_parameters import PRECISION


def check_all_state(res, res_box):
    if len(res) != len(res_box):
        return False

    for index in range(len(res)):
        amp_str = res[index]
        amp_str_box = res_box[index]
        real_str, image_str = amp_str.split(',')
        real_str_box, image_str_box = amp_str_box.split(',')
        test = float(image_str) - float(image_str_box)
        test1 = math.fabs(test)
        a = test1 > PRECISION
        if (math.fabs(float(real_str) - float(real_str_box)) > PRECISION
                or math.fabs(float(image_str) - float(image_str_box)) > PRECISION):
            return False

    return True


def test_cp_inverse_gate():
    # 使用本地量子计算模拟器
    circuit = QCircuit()
    qr = circuit.allocate(2)
    # 获取原始数据
    org_res = circuit.get_all_state()

    # 进行逆操作
    CP(pi / 2) * (qr[0], qr[1])
    CP(pi / 2) * (qr[0], qr[1])
    circuit.cmds[1].inverse = True

    # 获取逆操作后数据
    final_res = circuit.get_all_state()

    # 检查逆操作前后数据是否一致
    assert check_all_state(org_res, final_res)


if __name__ == "__main__":
    """运行test文件"""
    pytest.main(["-v", "-s", "./test_cp_inverse_gate.py"])


import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class CYGate(BasicGate):
    """Control Y gate."""

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "CY"

    def __or__(self, qubits):
        """
        Quantum logic gate operation

        Args:s
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.

        Example:
            CY * (qr[0], qr[1])
        """
        if len(qubits) != 2:
            raise AttributeError("参数错误：应该传入1个控制位，1个目标位")
        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(self, targets, controls, inverse=self.is_inverse)
        qubits[0].circuit.append_cmd(cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.array(
            [[0, 0, -1j, 0], [0, 1, 0, 0], [1j, 0, 0, 0], [0, 0, 0, 1]]
        )


CY = CYGate()



from qutrunk.converters.circuit_to_dag import circuit_to_dag


def _get_instructions(circuit):
    """
    Given a circuit, return a tuple(qubits, cbits, nodes)
    Args:
        circuit: a quantum circuit.
    Returns:
        qubits: a list of QuBit.
        cbits: a list of CBit.
        nodes: nodes is a list of DAG nodes whose type is "operation".
            (qubits, cbits, nodes)
    """
    # 1 将circuit表示为DAG
    dag = circuit_to_dag(circuit)

    # 2 获取量子电路的量子比特数qubits和经典比特数cbits
    qubits = dag.qubits
    cbits = dag.cbits

    # 3 获取DAG的操作节点
    nodes = []
    # 按照有向无环图的拓扑排序顺序返回节点
    for node in dag.topological_op_nodes():
        nodes.append([node])
    nodes = [[node for node in layer if any(q in qubits for q in node.qargs)] for layer in nodes]

    return qubits, cbits, nodes



"""
Contains ops for qutrunk.
"""
from .operation import Observable, Pauli, PauliArrary, PauliType, PauliCoeffi
from .paulii import PauliI
from .paulix import PauliX
from .pauliy import PauliY
from .pauliz import PauliZ



"""Command Module"""


class Command:
    """
    Converts the quantum gate operation into a specific command.

    Attributes:
        gate: quantum gate operator
        targets: target qubits
        controls: control qubits
        rotation: angle for rotation gate
    """

    def __init__(self, gate, targets, controls=[], rotation=[], inverse=False):
        # todo controls和rotation改成元组
        self.gate = gate
        self.controls = controls
        self.targets = targets
        self.rotation = rotation
        self.cmd_ver = "1.0"
        self.inverse = inverse

    def __eq__(self, other):
        """
        Two command are the same if they have teh same qasm.
        """
        if type(self) is not type(other):
            return False

        if self.qasm() == other.qasm():
            return True
        return False

    def __repr__(self) -> str:
        """Generate a representation of the command object instance

        Returns:
            str: A representation of the command instance
        """
        return f"Command(gate={self.gate}, controls={self.controls}, targets={self.targets}, rotation={self.rotation}), inverse={self.inverse})"

    def qasm(self):
        """
        generate OpenQASM code for command
        """
        name = str(self.gate).lower()

        if str(self.gate) == "Measure":
            index = self.targets[0]
            return f"{name} q[{str(index)}] -> c[{str(index)}]"

        ctrl_cnt = len(self.controls)
        if ctrl_cnt == 2:
            name = name.replace("c", "cc", 1)
        elif ctrl_cnt > 2:
            name = name.replace("c", "c" + str(ctrl_cnt), 1)

        angles_str = ""
        angles = self.gate.angles()
        if angles:
            angles_str = "(" + ",".join([str(ag) for ag in angles]) + ")"

        qubits_index = self.controls + self.targets
        qubits_str = ",".join([f"q[{qi}]" for qi in qubits_index])
        return name + angles_str + " " + qubits_str

    def quqasm(self):
        """
        generate QuQASM code for command
        """
        name = str(self.gate)
        params = []
        param_str = ""

        ctrl_cnt = len(self.controls)
        if ctrl_cnt > 1:
            params.append(ctrl_cnt)

        angles = self.gate.angles()
        if angles:
            params += angles

        if params:
            param_str = "(" + ", ".join([str(param) for param in params]) + ")"

        qubits_index = self.controls + self.targets
        qubits_str = ", ".join([f"q[{qi}]" for qi in qubits_index])
        if len(qubits_index) > 1:
            qubits_str = '(' + qubits_str + ')'

        return name + param_str + " * " + qubits_str

    @property
    def name(self):
        """command name"""
        return self.gate

    @property
    def num_qubits(self):
        """Return the number of qubits."""
        return len(self.controls) + len(self.targets)



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicRotateGate


class CU(BasicRotateGate):
    """
    Control U gate
    """

    def __init__(self, theta, phi, lam, gamma):
        """
        Args:
            theta: U gate parameter 1
            phi: U gate parameter 2
            lam:U gate parameter 3
            gamma: U gate parameter 4
        """
        super().__init__()
        self.theta = theta
        self.phi = phi
        self.lam = lam
        self.gamma = gamma

    def __str__(self):
        return "CU"

    def __or__(self, qubits):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to aplly CU gate
            
        Example:
            CU(pi/2,pi/2,pi/2,pi/2) * (qr[0], qr[1])
        """

        if len(qubits) != 2:
            raise AttributeError("Parameter Error: qubits should be two")
        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(self, targets, controls,
                      inverse=self.is_inverse, rotation=[self.theta, self.phi, self.lam, self.gamma])
        qubits[0].circuit.append_cmd(cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        half_theta = float(self.theta)
        half_phi = float(self.phi)
        half_lam = float(self.lam)
        half_gamma = float(self.gamma)
        cos = np.cos(half_theta / 2)
        sin = np.sin(half_theta / 2)
        a = np.exp(1j * half_gamma) * cos
        b = -np.exp(1j * (half_gamma + half_lam)) * sin
        c = np.exp(1j * (half_gamma + half_phi)) * sin
        d = np.exp(1j * (half_gamma + half_phi + half_lam)) * cos
        return np.array(
            [
                [1, 0, 0, 0],
                [0, a, 0, b],
                [0, 0, 1, 0],
                [0, c, 0, d],
            ]
        )



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class U2(BasicGate):
    """
    U2 gate
    """

    def __init__(self, theta, phi):
        """
        Args:
            theta: U2 gate parameter1
            phi: U2 gate parameter2
        """
        super().__init__()
        self.theta = theta
        self.phi = phi

    def __str__(self):
        return "U2"

    def __or__(self, qubit):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to apply U2 gate.

        Example:
            U2(pi/2, pi/2) * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.theta, self.phi], inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        isqrt2 = 1 / np.sqrt(2)
        phi = self.theta
        lam = self.phi
        return np.matrix(
            [
                [isqrt2, -np.exp(1j * lam) * isqrt2],
                [np.exp(1j * phi) * isqrt2, np.exp(1j * (phi + lam)) * isqrt2],
            ]
        )



"""run quantum program on IBM QE cloud platform."""
import uuid
import signal
import time
import requests

from requests import Session
from requests.compat import urljoin


# TODO:抽取到配置目录中
_AUTH_API_URL = 'https://auth.quantum-computing.ibm.com/api/users/loginWithToken'
_API_URL = 'https://api.quantum-computing.ibm.com/api/'

# TODO: call to get the API version automatically
CLIENT_APPLICATION = 'ibmqprovider/0.4.4'


class IBMQ(Session):
    """Manage a session between QuTrunk and the IBMQ web API."""

    def __init__(self, **kwargs):
        """Initialize a session with the IBM QE's APIs."""
        super().__init__(**kwargs)
        # 后端
        self.backends = {}
        # 超时时间
        self.timeout = 5.0

    def get_list_devices(self, verbose=False):
        """
        Get the list of available IBM backends with their properties.

        Args:
            verbose (bool): print the returned dictionary if True

        Returns:
            (dict) backends dictionary by name device, containing the qubit size 'nq',
            the coupling map 'coupling_map' as well as the device version 'version'
        """
        list_device_url = 'Network/ibm-q/Groups/open/Projects/main/devices/v/1'
        argument = {'allow_redirects': True, 'timeout': (self.timeout, None)}
        request = super().get(urljoin(_API_URL, list_device_url), **argument)
        # 抛出异常
        request.raise_for_status()
        # 反序列化
        r_json = request.json()
        # 重置backends
        self.backends = {}
        # 遍历返回结果
        for obj in r_json:
            self.backends[obj['backend_name']] = {
                'nq': obj['n_qubits'],
                'coupling_map': obj['coupling_map'],
                'version': obj['backend_version'],
            }
        # 如果verbose为True，打印backends
        if verbose:
            print('- List of IBMQ devices available:')
            print(self.backends)
        return self.backends

    def authenticate(self, token=None):
        """
        Authenticate with IBM's Web API.

        Args:
            token (str): IBM quantum experience user API token.
        """
        if token is None:
            raise Exception("please input the IBM QE token")
        # 将token写入headers
        self.headers.update({"X-Qx-Client-Application": CLIENT_APPLICATION})
        args = {
            "data": None,
            "json": {"apiToken": token},
            "timeout": (self.timeout, None),
        }
        request = super().post(_AUTH_API_URL, **args)
        # 抛出异常
        request.raise_for_status()
        # 反序列化
        r_json = request.json()
        self.params.update({'access_token': r_json['id']})

    def is_online(self, device):
        """设备是否可用"""
        return device in self.backends

    def can_run_experiment(self, info=None, device=""):
        """资源是否足够运行量子程序"""
        nb_qubit_max = self.backends[device]["nq"]
        nb_qubit_needed = info["nq"]
        return nb_qubit_needed <= nb_qubit_max, nb_qubit_max, nb_qubit_needed

    def run(self, info=None, device=""):
        """
        Run the quantum code to the IBMQ machine.

         Args:
            info (dict): dictionary sent by the backend containing the code to run
            device (str): name of the ibm device to use

        Returns:
            Execution Id
        """
        # 1 获取执行量子程序的设备URL和执行ID
        json_step1 = {
            'data': None,
            'json': {
                'backend': {'name': device},
                'allowObjectStorage': True,
                'shareLevel': 'none',
            },
            'timeout': (self.timeout, None),
        }
        request = super().post(
            urljoin(_API_URL, 'Network/ibm-q/Groups/open/Projects/main/Jobs'),
            **json_step1,
        )
        request.raise_for_status()
        r_json = request.json()
        print("r_json==", r_json)
        upload_url = r_json['objectStorageInfo']['uploadUrl']
        execution_id = r_json['id']

        # 2 上传circuit
        n_classical_reg = info['nq']
        n_qubits = n_classical_reg
        # 指令
        instructions = info['json']
        maxcredit = info['maxCredits']
        c_label = [["c", i] for i in range(n_classical_reg)]
        q_label = [["q", i] for i in range(n_qubits)]

        instruction_str = str(instructions).replace('\'', '\"')
        data = '{"qobj_id": "' + str(uuid.uuid4()) + '", '
        data += '"header": {"backend_name": "' + device + '", '
        data += '"backend_version": "' + self.backends[device]['version'] + '"}, '
        data += '"config": {"shots": ' + str(info['shots']) + ', '
        data += '"max_credits": ' + str(maxcredit) + ', "memory": false, '
        data += '"parameter_binds": [], "memory_slots": ' + str(n_classical_reg)
        data += ', "n_qubits": ' + str(n_qubits) + '}, "schema_version": "1.2.0", '
        data += '"type": "QASM", "experiments": [{"config": '
        data += '{"n_qubits": ' + str(n_qubits) + ', '
        data += '"memory_slots": ' + str(n_classical_reg) + '}, '
        data += '"header": {"qubit_labels": ' + str(q_label).replace('\'', '\"') + ', '
        data += '"n_qubits": ' + str(n_classical_reg) + ', '
        data += '"qreg_sizes": [["q", ' + str(n_qubits) + ']], '
        data += '"clbit_labels": ' + str(c_label).replace('\'', '\"') + ', '
        data += '"memory_slots": ' + str(n_classical_reg) + ', '
        data += '"creg_sizes": [["c", ' + str(n_classical_reg) + ']], '
        data += '"name": "circuit0", "global_phase": 0}, "instructions": ' + instruction_str + '}]}'

        json_step2 = {
            'data': data,
            'params': {'access_token': None},
            'timeout': (5.0, None),
        }
        request = super().put(upload_url, **json_step2)
        request.raise_for_status()

        # 3 确认上传
        json_step3 = {'data': None, 'json': None, 'timeout': (self.timeout, None)}

        upload_data_url = urljoin(
            _API_URL,
            'Network/ibm-q/Groups/open/Projects/main/Jobs/' + str(execution_id) + '/jobDataUploaded',
        )
        request = super().post(upload_data_url, **json_step3)
        request.raise_for_status()

        return execution_id

    def get_result(self, device, execution_id, num_retries=3000, interval=1, verbose=False):
        """获取执行结果"""
        job_status_url = 'Network/ibm-q/Groups/open/Projects/main/Jobs/' + execution_id
        if verbose:
            print(f"Waiting for results. [Job ID: {execution_id}]")
            original_sigint_handler = signal.getsignal(signal.SIGINT)
            print("signal=", original_sigint_handler)

            def _handle_sigint_during_get_result(*_):
                raise Exception("Interrupted. The ID of your submitted job is {}.".format(execution_id))

            try:
                signal.signal(signal.SIGINT, _handle_sigint_during_get_result)

                for retries in range(num_retries):
                    # 1: 等待job的运行
                    json_step5 = {'allow_redirects': True, 'timeout': (self.timeout, None)}
                    request = super().get(urljoin(_API_URL, job_status_url), **json_step5)
                    request.raise_for_status()
                    r_json = request.json()
                    acceptable_status = ['VALIDATING', 'VALIDATED', 'RUNNING']

                    # 2 获取执行结果
                    if r_json['status'] == 'COMPLETED':
                        json_step6 = {
                            'allow_redirects': True,
                            'timeout': (self.timeout, None),
                        }
                        request = super().get(
                            urljoin(_API_URL, job_status_url + '/resultDownloadUrl'),
                            **json_step6,
                        )
                        request.raise_for_status()
                        r_json = request.json()

                        json_step7 = {
                            'allow_redirects': True,
                            'params': {'access_token': None},
                            'timeout': (self.timeout, None),
                        }
                        request = super().get(r_json['url'], **json_step7)
                        r_json = request.json()
                        result = r_json['results'][0]

                        json_step8 = {'data': None, 'json': None, 'timeout': (5.0, None)}
                        request = super().post(
                            urljoin(_API_URL, job_status_url + '/resultDownloaded'),
                            **json_step8,
                        )
                        r_json = request.json()  # {'terminated': True}

                        return result

                    if r_json['status'] not in acceptable_status:
                        raise Exception(f"Error while running the code. Last status: {r_json['status']}.")
                    time.sleep(interval)
                    if self.is_online(device) and retries % 60 == 0:
                        self.get_list_devices()
                        if not self.is_online(device):
                            raise Exception(
                                f"Device went offline. The ID of your submitted job is {execution_id}."
                            )

            finally:
                if original_sigint_handler is not None:
                    signal.signal(signal.SIGINT, original_sigint_handler)

            raise Exception(f"Timeout. The ID of your submitted job is {execution_id}.")


def show_devices(token=None, verbose=False):
    """
    Access the list of available devices.

    Args:
        token (str): IBM quantum experience user API token.
        verbose (bool): If True, additional information is printed

    Returns:
        (list) list of available devices and their properties
    """
    ibmq_session = IBMQ()
    ibmq_session.authenticate(token=token)
    return ibmq_session.get_list_devices(verbose=verbose)


def send(info, device='ibmq_qasm_simulator', token=None, shots=None, num_retries=3000,
         interval=1, verbose=False):
    """
    Send QASM through the IBM API and runs the quantum circuit.

    Args:
        info(dict): Contains representation of the circuit to run.
        device (str): name of the ibm device. Simulator chosen by default
        token (str): IBM quantum experience user API token.
        shots (int): Number of runs of the same circuit to collect statistics.
        verbose (bool): If True, additional information is printed, such as measurement statistics. Otherwise, the
            backend simply registers one measurement result (same behavior as the projectq Simulator).

    Returns:
        (dict) result form the IBMQ server

    """
    try:
        ibmq_session = IBMQ()
        if shots is not None:
            info['shots'] = shots
        if verbose:
            print("- Authenticating...")
            if token is not None:
                print('user API token: ' + token)
        ibmq_session.authenticate(token)

        # check if the device is online
        ibmq_session.get_list_devices(verbose)
        # 设备是否可用
        online = ibmq_session.is_online(device)
        if not online:
            print("The device is offline. Use the simulator instead or try again later.")
            raise Exception("Device is offline.")

        # check if the device has enough qubit to run the code
        # 资源是否够用
        runnable, qmax, qneeded = ibmq_session.can_run_experiment(info, device)
        if not runnable:
            print(
                    f"The device is too small ({qmax} qubits available) for the code "
                    + f"requested({qneeded} qubits needed) Try to look for another "
                    + "device with more qubits"
            )
            raise Exception("Device is too small.")
        if verbose:
            print(f"- Running code: {info}")
        execution_id = ibmq_session.run(info, device)
        if verbose:
            print("- Waiting for results...")
        res = ibmq_session.get_result(
            device,
            execution_id,
            num_retries=num_retries,
            interval=interval,
            verbose=verbose,
        )
        if verbose:
            print("- Done.")
        return res
    except requests.exceptions.HTTPError as err:
        print("- There was an error running your code:")
        print(err)
    except requests.exceptions.RequestException as err:
        print("- Looks like something is wrong with server:")
        print(err)
    except KeyError as err:
        print("- Failed to parse response:")
        print(err)
    return None


if __name__ == '__main__':
    # i = IBMQ()
    # token = "247891ade16963963eb432d5ae4c7bbd1948d893f256a9f5d94af94628c5b57c73877dbf6ad4d2bd0ffc0e61d6aa001897666f1d75c3e613784ed8f2c7cafe68"
    #
    # print(i.authenticate(token))
    # i.get_list_devices(verbose=True)
    # print(i.is_online("ibmq_qasm_simulator"))
    # i.run(device="ibmq_qasm_simulator")

    # 获取设备列表
    token = "247891ade16963963eb432d5ae4c7bbd1948d893f256a9f5d94af94628c5b57c73877dbf6ad4d2bd0ffc0e61d6aa001897666f1d75c3e613784ed8f2c7cafe68"
    show_devices(token, verbose=True)



import abc
from enum import Enum
from typing import List, Union


class PauliType(Enum):
    """PauliType"""
    POT_PAULI_I = 0
    POT_PAULI_X = 1
    POT_PAULI_Y = 2
    POT_PAULI_Z = 3


class Observable(abc.ABC):
    """Observable class"""
    @abc.abstractmethod
    def obs_data(self):
        """return Observable data."""


class Pauli(Observable):
    """Pauli operator base class"""
    def __init__(self, targets: List[int], pauli_type=PauliType.POT_PAULI_I):
        self.targets = targets
        self.pauli_type = pauli_type

    def obs_data(self):
        """Get Observable data."""
        puali_list = []
        for temp in self.targets:
            pauli = {}
            pauli['oper_type'] = self.pauli_type.value
            pauli['target'] = temp
            puali_list.append(pauli)
        return puali_list


class PauliArrary(Observable):
    """Apply the Pauli"""

    def __init__(self, obj_arrary: List[Pauli]):
        self.obj_arrary = obj_arrary

    def obs_data(self):
        """Get Observable data."""
        puali_list = []
        for obj in self.obj_arrary:
            puali_list = puali_list + obj.obs_data()
        return puali_list


class PauliCoeffi(Observable):
    """Apply the PauliCoeffi"""
    def __init__(self, pauli_types: List[PauliType], coeffis: List[float]):
        self.pauli_types = pauli_types
        self.coeffis = coeffis

    def obs_data(self):
        """Get Observable data."""
        paulis = []
        for temp in self.pauli_types:
            paulis.append(temp.value)
        return (paulis, self.coeffis)



import sys
sys.path.append("D:/code/qike/qubox/src") 
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, All, iSwap
from qutrunk.backends import BackendQuSprout, ExecType
from numpy import pi

def testiswap(backend=None):
    # allocate
    qc = QCircuit(backend=backend)
    qureg = qc.allocate(2)
    H | qureg[0]
    H | qureg[1]
    iSwap(pi/2) | (qureg[0], qureg[1])

    #qc, qureg = qc.inverse()

    All(Measure) | qureg
    qc.print()
    result = qc.run()

    print("执行iswap，测量得到的结果是：", result.get_measure()[0])
    return qc

if __name__ == "__main__":
    qc = testiswap()
    print(qc.draw())



import sys

sys.path.append("D:/code/qike/qubox/src")
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, All, CR
from qutrunk.backends import BackendQuSprout, ExecType
from numpy import pi


def testcr(backend=None):
    # allocate
    qc = QCircuit(backend=backend)
    qu_reg = qc.allocate(2)
    H * qureg[0]
    H * qureg[1]
    CR(pi / 2) * (qureg[0], qureg[1])

    qc, qureg = qc.inverse()

    All(Measure) * qureg
    qc.print()
    result = qc.run()

    print("执行CR，测量得到的结果是：", result.get_measure()[0])
    return qc


if __name__ == "__main__":
    qc = testcr()
    print(qc.draw())



import math

import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicRotateGate


class Rx(BasicRotateGate):
    """
    Rotate a single qubit by a given angle around the X-axis of the Bloch-sphere

    Args:
        alpha: the angle to rotate
    """

    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Rx"

    def __or__(self, qubit):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to aplly Rx gate。

        Example:
            Rx(alpha) * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.rotation], inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [math.cos(0.5 * self.rotation), -1j * math.sin(0.5 * self.rotation)],
                [-1j * math.sin(0.5 * self.rotation), math.cos(0.5 * self.rotation)],
            ]
        )



import cmath

import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicRotateGate


class Ryy(BasicRotateGate):
    """
    RotationYY gate class

    Args:
        alpha: the angle to rotate
    """

    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "Ryy"

    def __or__(self, qubits):
        """
        Quantum logic gate operation

        Args:
            qubits: the quantum bits to apply Ryy gate.

        Example:
            Ryy(alpha) * (qr[0], qr[1])
        """
        if len(qubits) != 2:
            raise AttributeError()
        targets = [q.index for q in qubits]
        cmd = Command(self, targets, rotation=[self.rotation], inverse=self.is_inverse)
        qubits[0].circuit.append_cmd(cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix(
            [
                [cmath.cos(0.5 * self.rotation), 0, 0, 1j * cmath.sin(0.5 * self.rotation)],
                [0, cmath.cos(0.5 * self.rotation), -1j * cmath.sin(0.5 * self.rotation), 0],
                [0, -1j * cmath.sin(0.5 * self.rotation), cmath.cos(0.5 * self.rotation), 0],
                [1j * cmath.sin(0.5 * self.rotation), 0, 0, cmath.cos(0.5 * self.rotation)],
            ]
        )



"""Hadamard gate."""

import cmath

import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class HGate(BasicGate):
    """
    Apply the single-qubit Hadamard gate
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "H"

    def __or__(self, qubit):
        """
        Quantum logic gate operation
        
        Args:
            qubit: quantum bit applied to hadamard gate
        Example:
            H * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return 1.0 / cmath.sqrt(2.0) * np.matrix([[1, 1], [1, -1]])

    @property
    def label(self):
        """a label for identifying the gate."""
        self.__str__()


H = HGate()



import math

import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicRotateGate


class R(BasicRotateGate):
    """
    Rotation alpha around the cos(beta)x + sin(beta)y axis

    Args:
        theta: the angle to rotate
        phi: define the axis cos(beta)x + sin(beta)y
    """

    def __init__(self, theta, phi):
        super().__init__()
        self.theta = theta
        self.phi = phi

    def __str__(self):
        return "R"

    def __or__(self, qubit):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to aplly R gate.

        Example:
            R(theta, phi) * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.theta, self.phi], inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        theta, phi = float(self.theta), float(self.phi)
        cos = math.cos(theta / 2)
        sin = math.sin(theta / 2)
        exp_m = np.exp(-1j * phi)
        exp_p = np.exp(1j * phi)
        return np.array([[cos, -1j * exp_m * sin], [-1j * exp_p * sin, cos]])



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate


class XGate(BasicGate):
    """Apply the single-qubit Pauli-X (also known as the X, sigma-X, NOT or bit-flip) gate"""

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "X"

    def __or__(self, qubit):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to aplly X gate.

        Example:
            X * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[0, 1], [1, 0]])


X = NOT = XGate()



import pytest
import json

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, S, X, U1, U2, U3
from numpy import pi


def test_u1_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    U1(pi / 2) | qr[0]
    # S | qr[0]
    target = '["real, imag", "1.000000000000, 0.000000000000", "0.000000000000, 0.000000000000"]'
    res = json.dumps(circuit.get_all_state())
    # print(circuit.get_all_state())
    assert res == target
    circuit.run()


def test_u2_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    U2(0, pi) | qr[0]
    # H | qr[0]
    target = '["real, imag", "0.707106781187, 0.000000000000", "0.707106781187, 0.000000000000"]'
    res = json.dumps(circuit.get_all_state())
    # print(circuit.get_all_state()
    assert res == target
    circuit.run()


def test_u3_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    U3(pi, 0, pi) | qr[0]
    # X | qr[0]
    target = '["real, imag", "0.000000000000, 0.000000000000", "1.000000000000, 0.000000000000"]'
    res = json.dumps(circuit.get_all_state())
    # print(circuit.get_all_state())
    assert res == target
    circuit.run()



# TODO: 用于统计量子线路所需的资源：量子比特数，量子门数；量子线路编译耗时，运行耗时
import time


class Counter:
    """
    resources of circuit
    """
    def __init__(self, circuit):
        self.circuit = circuit
        self.qubits = 0
        self.quantum_gates = 0
        self.start_time = time.time()
        self.total_time = 0
        self.backend_time = 0
        self.qutrunk_time = 0

    def acc_run_time(self, elapsed):
        """Accumulate backend running time"""
        self.backend_time += elapsed

    def finish(self):
        """Statistics time and gates when circuit running finish"""
        self.total_time = time.time() - self.start_time
        self.qutrunk_time = self.total_time - self.backend_time
        self.quantum_gates = self.circuit.gates_len
    
    def __repr__(self):
        return f"Counter(quit={self.qubits})"

    def show_verbose(self):
        """Print the resource info of circuit"""
        print("==================Counter==================")
        print(self)
        print("qubits =", self.qubits)
        print("quantum_gates =", self.quantum_gates)
        print("total_time =", self.total_time)
        print("qutrunk_time =", self.qutrunk_time)
        print("backend_time =", self.backend_time)



from qutrunk.circuit.ops.operation import Pauli, PauliType


class PauliX(Pauli):
    """Apply the PauliX"""

    def __init__(self, *targets):
        super().__init__(list(targets), PauliType.POT_PAULI_X)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qutrunk',
 'qutrunk.backends',
 'qutrunk.backends.ibm',
 'qutrunk.circuit',
 'qutrunk.circuit.gates',
 'qutrunk.circuit.ops',
 'qutrunk.converters',
 'qutrunk.dagcircuit',
 'qutrunk.example',
 'qutrunk.qasm',
 'qutrunk.qasm.node',
 'qutrunk.sim',
 'qutrunk.sim.local',
 'qutrunk.sim.local.pybind11',
 'qutrunk.sim.local.pybind11.docs',
 'qutrunk.sim.local.pybind11.pybind11',
 'qutrunk.sim.local.pybind11.tests',
 'qutrunk.sim.local.pybind11.tests.extra_python_package',
 'qutrunk.sim.local.pybind11.tests.extra_setuptools',
 'qutrunk.sim.local.pybind11.tests.test_cmake_build',
 'qutrunk.sim.local.pybind11.tests.test_embed',
 'qutrunk.sim.local.pybind11.tools',
 'qutrunk.sim.qusprout',
 'qutrunk.sim.qusprout.code',
 'qutrunk.sim.qusprout.qusprout',
 'qutrunk.sim.qusprout.qusproutdata',
 'qutrunk.sim.qusprout.work',
 'qutrunk.test',
 'qutrunk.test.gate',
 'qutrunk.tools',
 'qutrunk.visualizations']

package_data = \
{'': ['*'],
 'qutrunk': ['config/*',
             'doc/distribute_package.md',
             'doc/distribute_package.md',
             'doc/project_documentation.md',
             'doc/project_documentation.md'],
 'qutrunk.qasm': ['libs/*'],
 'qutrunk.sim': ['Release/*',
                 'local/QuEST/*',
                 'local/QuEST/include/*',
                 'local/QuEST/src/*',
                 'local/QuEST/src/CPU/*',
                 'local/QuEST/src/GPU/*',
                 'local/pybind11/docs/_static/*',
                 'local/pybind11/docs/advanced/*',
                 'local/pybind11/docs/advanced/cast/*',
                 'local/pybind11/docs/advanced/pycpp/*',
                 'local/pybind11/docs/cmake/*',
                 'local/pybind11/include/pybind11/*',
                 'local/pybind11/include/pybind11/detail/*',
                 'local/pybind11/include/pybind11/stl/*',
                 'local/pybind11/tests/test_cmake_build/installed_embed/*',
                 'local/pybind11/tests/test_cmake_build/installed_function/*',
                 'local/pybind11/tests/test_cmake_build/installed_target/*',
                 'local/pybind11/tests/test_cmake_build/subdirectory_embed/*',
                 'local/pybind11/tests/test_cmake_build/subdirectory_function/*',
                 'local/pybind11/tests/test_cmake_build/subdirectory_target/*']}

install_requires = \
['networkx>=2.8,<3.0',
 'numba>=0.56.0,<0.57.0',
 'numpy>=1.22.3,<2.0.0',
 'ply>=3.11,<4.0',
 'pyyaml>=6.0,<7.0',
 'requests>=2.27.1,<3.0.0',
 'retworkx>=0.11.0,<0.12.0',
 'thrift>=0.15.0,<0.16.0']

setup_kwargs = {
    'name': 'qutrunk',
    'version': '0.1.9',
    'description': 'qutrunk is an open source library for quantum computing.',
    'long_description': '## qutrunk\nqutrunk is an quantum programming framework developed by QuDoor corporation, provide the required APIs for quantum programming, it works as follow: \n\nExport quantum programs based on quantum circuit, qubit and quantum gate, etc. Then you can run the \nquantum circuit by the following two ways:\n1. Use PC for quantum computing(work as the default way). \n2. Connect with QuBox(a quantum device developed by qudoor), to run quantum programs, it provide more powerful quantum computing. \n\n## Install\n1. Install from whl package, run the following command directly:\n```\npip install qutrunk\n```\n\n2. Install from source code, run the following command to install qutrunk by source code:\n```\npython3 setup.py install\n```\n\n## Example:\nbell-pair quantum algorithm：\n\n```\nfrom qutrunk.circuit import QCircuit\nfrom qutrunk.circuit.gates import H, CNOT, Measure\n\nqc = QCircuit()\nqr = qc.allocate(2) # allocate\n\nH | qr[0]   # apply gate\nCNOT | (qr[0], qr[1])\nMeasure | qr[0]\nMeasure | qr[1]\n\nqc.print(qc)   # print circuit\nres = qc.run(shots=1024) # run circuit\nprint(res.get_counts()) # print result\n```\n\n ## License\n Apache Licence 2.0\n\n Copyright (c) 2022-2025 By QUDOOR Technolgies Inc. All Right Reserved.',
    'author': 'qudoorzh2022',
    'author_email': 'qudoorzh2022@163.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://www.qudoor.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



import yaml
import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_qubox_setting():
    """Get Qubox configure info"""
    yaml_file = os.path.join(BASE_DIR, 'config', 'qubox.yaml')  # yaml文件路径
    with open(yaml_file) as f:
        yaml_content = yaml.load(f, Loader=yaml.FullLoader)
        return "{ip}:{port}".format(ip=yaml_content["ip"], port=yaml_content["port"])

        
print(get_qubox_setting())


import cmath

import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasciPhaseGate


class P(BasciPhaseGate):
    """
    Phase gate (global phase)

    Args:
        alpha: the phase to apply
    """

    def __init__(self, alpha):
        super().__init__()
        self.rotation = alpha

    def __str__(self):
        return "P"

    def __or__(self, qubit):
        """
        Quantum logic gate operation

        Args:
            qubit: the quantum bit to aplly phase gate.

        Example:
            P(alpha) * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, rotation=[self.rotation], inverse=self.is_inverse)
        qubit.circuit.append_cmd(cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement"""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[cmath.exp(1j * self.rotation), 0], [0, cmath.exp(1j * self.rotation)]])



import pytest
import json

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import S, X, U1, U2, U3, CZ,\
    H, Measure, CNOT, Toffoli, P, R, Rx, Ry, Rz, S, Sdg, T, Tdg, X, Y, Z, MCX, MCZ,\
    NOT, Swap, SqrtSwap, SqrtX, All, CP, CX, CY, CZ, CRx, CRy, CRz, Rxx, Ryy, Rzz, \
    U1, U2, U3, Barrier, iSwap, CR, CU, CU1, CU3, X1, Y1

from numpy import pi
from qutrunk.backends import BackendQuSprout, backend
import math

PRECISION = 0.0000000001

def check_all_state(res, resbox):
    if len(res) != len(resbox):
        return False

    for index in range(len(res)):
        ampstr = res[index]
        ampstrbox = resbox[index]
        realstr, imagstr = ampstr.split(',')
        realstrbox, imagstrbox = ampstrbox.split(',')
        if (math.fabs(float(realstr) - float(realstrbox)) > PRECISION 
            or math.fabs(float(imagstr) - float(imagstrbox) > PRECISION)):
            return False
    
    return True

def test_h_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    H | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    H | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_p_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    P(pi/2) | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    P(pi/2) | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cp_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CP(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CP(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_r_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    R(pi/2, pi/2) | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    R(pi/2, pi/2) | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_rx_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    Rx(pi/2) | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    Rx(pi/2) | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_rxx_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    Rxx(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    Rxx(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_ryy_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    Ryy(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    Ryy(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_rzz_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    Rzz(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    Rzz(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_not_gate():
    test_x_gate()

def test_x_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    X | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    X | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_y_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    Y | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    Y | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_z_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    Z | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    Z | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_s_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    S | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    S | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_t_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    T | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    T | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_sdg_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    Sdg | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    Sdg | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_tdg_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    Tdg | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    Tdg | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_sqrtswap_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    SqrtSwap | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    SqrtSwap | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_swap_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    Swap | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    Swap | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cnot_gate():
    test_cx_gate()

def test_cx_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CX | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CX | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cy_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CY | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CY | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cz_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CZ | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CZ | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_u3_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    U3(pi, 0, pi) | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    U3(pi, 0, pi) | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_u2_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    U2(0, pi) | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    U2(0, pi) | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_u1_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    U1(pi/2) | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    U1(pi/2) | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_crx_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CRx(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CRx(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cry_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CRy(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CRy(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_crz_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CRz(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CRz(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_x1_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    X1 | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    X1 | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_y1_gate():
    circuit = QCircuit()
    qr = circuit.allocate(1)
    Y1 | qr[0]
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(1)
    Y1 | qrbox[0]
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cu1_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CU1(pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CU1(pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cu3_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CU3(pi/2,pi/2,pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CU3(pi/2,pi/2,pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_u_gate():
    test_u3_gate()

def test_cu_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CU(pi/2,pi/2,pi/2,pi/2) | (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CU(pi/2,pi/2,pi/2,pi/2) | (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_cr_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    CR(pi/2) * (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    CR(pi/2) * (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)

def test_iswap_gate():
    circuit = QCircuit()
    qr = circuit.allocate(2)
    iSwap(pi/2) * (qr[0], qr[1])
    res = circuit.get_all_state()

    circuitbox = QCircuit(backend=BackendQuSprout())
    qrbox = circuitbox.allocate(2)
    iSwap(pi/2) * (qrbox[0], qrbox[1])
    resbox = circuitbox.get_all_state()

    assert check_all_state(res, resbox)